import sys
import numpy as np
from phonopy.phonon.degeneracy import degenerate_sets
from phonopy.harmonic.force_constants import similarity_transformation
from phonopy.units import THz, Angstrom
from phono3py.phonon3.conductivity import (Conductivity, all_bands_exist,
                                           unit_to_WmK)
from phono3py.phonon3.collision_matrix import CollisionMatrix
from phono3py.phonon3.triplets import (get_grid_points_by_rotations,
                                       get_BZ_grid_points_by_rotations)
from phono3py.file_IO import (write_kappa_to_hdf5,
                              write_collision_to_hdf5,
                              read_collision_from_hdf5,
                              write_collision_eigenvalues_to_hdf5,
                              write_unitary_matrix_to_hdf5)
from phonopy.units import THzToEv, Kb

def get_thermal_conductivity_LBTE(
        interaction,
        symmetry,
        temperatures=np.arange(0, 1001, 10, dtype='double'),
        sigmas=None,
        sigma_cutoff=None,
        is_isotope=False,
        mass_variances=None,
        grid_points=None,
        boundary_mfp=None, # in micrometre
        is_reducible_collision_matrix=False,
        is_kappa_star=True,
        gv_delta_q=1e-4, # for group velocity
        is_full_pp=False,
        pinv_cutoff=1.0e-8,
        pinv_solver=0, # default: dsyev in lapacke
        write_collision=False,
        read_collision=False,
        write_kappa=False,
        input_filename=None,
        output_filename=None,
        log_level=0):

    if sigmas is None:
        sigmas = []
    if log_level:
        print("-------------------- Lattice thermal conducitivity (LBTE) "
              "--------------------")
        print("Cutoff frequency of pseudo inversion of collision matrix: %s" %
              pinv_cutoff)

    if read_collision:
        temps = None
    else:
        temps = temperatures

    lbte = Conductivity_LBTE(
        interaction,
        symmetry,
        grid_points=grid_points,
        temperatures=temps,
        sigmas=sigmas,
        sigma_cutoff=sigma_cutoff,
        is_isotope=is_isotope,
        mass_variances=mass_variances,
        boundary_mfp=boundary_mfp,
        is_reducible_collision_matrix=is_reducible_collision_matrix,
        is_kappa_star=is_kappa_star,
        gv_delta_q=gv_delta_q,
        is_full_pp=is_full_pp,
        pinv_cutoff=pinv_cutoff,
        pinv_solver=pinv_solver,
        log_level=log_level)

    if read_collision:
        read_from = _set_collision_from_file(
            lbte,
            indices=read_collision,
            is_reducible_collision_matrix=is_reducible_collision_matrix,
            filename=input_filename,
            log_level=log_level)
        if not read_from:
            print("Reading collision failed.")
            return False
        if log_level:
            temperatures = lbte.get_temperatures()
            if len(temperatures) > 5:
                text = (" %.1f " * 5 + "...") % tuple(temperatures[:5])
                text += " %.1f" % temperatures[-1]
            else:
                text = (" %.1f " * len(temperatures)) % tuple(temperatures)
            print("Temperature: " + text)

    for i in lbte:
        if write_collision:
            _write_collision(
                lbte,
                interaction,
                i=i,
                is_reducible_collision_matrix=is_reducible_collision_matrix,
                is_one_gp_colmat=(grid_points is not None),
                filename=output_filename)

    if (not read_collision and all_bands_exist(interaction)
        or (read_collision and read_from == "grid_points")):
        if grid_points is None:
            _write_collision(lbte, interaction, filename=output_filename)

    if write_kappa:
        if grid_points is None and all_bands_exist(interaction):
            lbte.set_kappa_at_sigmas()
            _write_kappa(
                lbte,
                interaction.get_primitive().get_volume(),
                is_reducible_collision_matrix=is_reducible_collision_matrix,
                pinv_solver=pinv_solver,
                filename=output_filename,
                log_level=log_level)

    return lbte

def _write_collision(lbte,
                     interaction,
                     i=None,
                     is_reducible_collision_matrix=False,
                     is_one_gp_colmat=False,
                     filename=None):
    grid_points = lbte.get_grid_points()
    temperatures = lbte.get_temperatures()
    sigmas = lbte.get_sigmas()
    sigma_cutoff = lbte.get_sigma_cutoff_width()
    gamma = lbte.get_gamma()
    gamma_isotope = lbte.get_gamma_isotope()
    collision_matrix = lbte.get_collision_matrix()
    mesh = lbte.get_mesh_numbers()

    print(collision_matrix.shape)

    if i is not None:
        gp = grid_points[i]
        if is_one_gp_colmat:
            igp = 0
        else:
            if is_reducible_collision_matrix:
                igp = gp
            else:
                igp = i
        if all_bands_exist(interaction):
            for j, sigma in enumerate(sigmas):
                if gamma_isotope is not None:
                    gamma_isotope_at_sigma = gamma_isotope[j, igp]
                else:
                    gamma_isotope_at_sigma = None
                write_collision_to_hdf5(
                    temperatures,
                    mesh,
                    gamma=gamma[j, :, igp],
                    gamma_isotope=gamma_isotope_at_sigma,
                    collision_matrix=collision_matrix[j, :, igp],
                    grid_point=gp,
                    sigma=sigma,
                    sigma_cutoff=sigma_cutoff,
                    filename=filename)
        else:
            for j, sigma in enumerate(sigmas):
                for k, bi in enumerate(interaction.get_band_indices()):
                    if gamma_isotope is not None:
                        gamma_isotope_at_sigma = gamma_isotope[j, igp, k]
                    else:
                        gamma_isotope_at_sigma = None
                    write_collision_to_hdf5(
                        temperatures,
                        mesh,
                        gamma=gamma[j, :, igp, k],
                        gamma_isotope=gamma_isotope_at_sigma,
                        collision_matrix=collision_matrix[j, :, igp, k],
                        grid_point=gp,
                        band_index=bi,
                        sigma=sigma,
                        sigma_cutoff=sigma_cutoff,
                        filename=filename)
    else:
        for j, sigma in enumerate(sigmas):
            if gamma_isotope is not None:
                gamma_isotope_at_sigma = gamma_isotope[j]
            else:
                gamma_isotope_at_sigma = None
            write_collision_to_hdf5(temperatures,
                                    mesh,
                                    gamma=gamma[j],
                                    gamma_isotope=gamma_isotope_at_sigma,
                                    collision_matrix=collision_matrix[j],
                                    sigma=sigma,
                                    sigma_cutoff=sigma_cutoff,
                                    filename=filename)

def _write_kappa(lbte,
                 volume,
                 is_reducible_collision_matrix=False,
                 pinv_solver=None,
                 filename=None,
                 log_level=0):
    temperatures = lbte.get_temperatures()
    sigmas = lbte.get_sigmas()
    sigma_cutoff = lbte.get_sigma_cutoff_width()
    mesh = lbte.get_mesh_numbers()
    weights = lbte.get_grid_weights()
    frequencies = lbte.get_frequencies()
    ave_pp = lbte.get_averaged_pp_interaction()
    qpoints = lbte.get_qpoints()
    kappa = lbte.get_kappa()
    kappa_RTA = lbte.get_kappa_RTA()
    coleigs = lbte.get_collision_eigenvalues()
    # After kappa calculation, the variable is overwritten by unitary matrix
    unitary_matrix = lbte.get_collision_matrix()

    if is_reducible_collision_matrix:
        ir_gp = lbte.get_grid_points()
        # frequencies = lbte.get_frequencies_all()
        gamma = lbte.get_gamma()[:, :, ir_gp, :]
        gamma_isotope = lbte.get_gamma_isotope()
        if gamma_isotope is not None:
            gamma_isotope = gamma_isotope[:, ir_gp, :]
        gv = lbte.get_group_velocities()[ir_gp]
        gv_by_gv = lbte.get_gv_by_gv()[ir_gp]
        mode_cv = lbte.get_mode_heat_capacities()[:, ir_gp, :]
        mode_kappa = lbte.get_mode_kappa()[:, :, ir_gp, :, :]
        mode_kappa_RTA = lbte.get_mode_kappa_RTA()[:, :, ir_gp, :, :]
        for i, w in enumerate(weights):
            mode_kappa[:, :, i, :, :] *= w
            mode_kappa_RTA[:, :, i, :, :] *= w
        mfp = lbte.get_mean_free_path()[:, :, ir_gp, :, :]
    else:
        frequencies = lbte.get_frequencies()
        gamma = lbte.get_gamma()
        gamma_isotope = lbte.get_gamma_isotope()
        gv = lbte.get_group_velocities()
        gv_by_gv = lbte.get_gv_by_gv()
        mode_cv = lbte.get_mode_heat_capacities()
        mode_kappa = lbte.get_mode_kappa()
        mode_kappa_RTA = lbte.get_mode_kappa_RTA()
        mfp = lbte.get_mean_free_path()

    for i, sigma in enumerate(sigmas):
        if gamma_isotope is not None:
            gamma_isotope_at_sigma = gamma_isotope[i]
        else:
            gamma_isotope_at_sigma = None
        write_kappa_to_hdf5(temperatures,
                            mesh,
                            frequency=frequencies,
                            group_velocity=gv,
                            gv_by_gv=gv_by_gv,
                            mean_free_path=mfp[i],
                            heat_capacity=mode_cv,
                            kappa=kappa[i],
                            mode_kappa=mode_kappa[i],
                            kappa_RTA=kappa_RTA[i],
                            mode_kappa_RTA=mode_kappa_RTA[i],
                            gamma=gamma[i],
                            gamma_isotope=gamma_isotope_at_sigma,
                            averaged_pp_interaction=ave_pp,
                            qpoint=qpoints,
                            weight=weights,
                            sigma=sigma,
                            sigma_cutoff=sigma_cutoff,
                            kappa_unit_conversion=unit_to_WmK / volume,
                            filename=filename,
                            verbose=log_level)

        if coleigs is not None:
            write_collision_eigenvalues_to_hdf5(temperatures,
                                                mesh,
                                                coleigs[i],
                                                sigma=sigma,
                                                sigma_cutoff=sigma_cutoff,
                                                filename=filename,
                                                verbose=log_level)
            if pinv_solver is not None:
                solver = pinv_solver
                if pinv_solver == 0:
                    try:
                        import scipy.linalg
                    except ImportError:
                        solver = 0
                    else:
                        solver = 6
                if solver % 10 == 6:
                    write_unitary_matrix_to_hdf5(temperatures,
                                                 mesh,
                                                 unitary_matrix=unitary_matrix,
                                                 sigma=sigma,
                                                 sigma_cutoff=sigma_cutoff,
                                                 filename=filename)

def _set_collision_from_file(lbte,
                             indices='all',
                             is_reducible_collision_matrix=False,
                             filename=None,
                             log_level=0):
    sigmas = lbte.get_sigmas()
    sigma_cutoff = lbte.get_sigma_cutoff_width()
    mesh = lbte.get_mesh_numbers()
    grid_points = lbte.get_grid_points()
    indices = indices

    if len(sigmas) > 1:
        gamma = []
        collision_matrix = []

    read_from = None

    if log_level:
        print("---------------------- Reading collision data from file "
              "----------------------")
        sys.stdout.flush()

    for j, sigma in enumerate(sigmas):
        collisions = read_collision_from_hdf5(mesh,
                                              indices=indices,
                                              sigma=sigma,
                                              sigma_cutoff=sigma_cutoff,
                                              filename=filename,
                                              verbose=(log_level > 0))
        if log_level:
            sys.stdout.flush()

        if collisions:
            (colmat_at_sigma,
             gamma_at_sigma,
             temperatures) = collisions

            if len(sigmas) == 1:
                collision_matrix = colmat_at_sigma
                gamma = np.zeros((1,) + gamma_at_sigma.shape,
                                 dtype='double', order='C')
                gamma[0] = gamma_at_sigma
            else:
                collision_matrix.append(colmat_at_sigma)
                gamma.append(gamma_at_sigma)
            read_from = "full_matrix"
        else:
            vals = _allocate_collision(True,
                                       mesh,
                                       sigma,
                                       sigma_cutoff,
                                       grid_points,
                                       indices,
                                       is_reducible_collision_matrix,
                                       filename)
            if vals:
                colmat_at_sigma, gamma_at_sigma, temperatures = vals
            else:
                if log_level:
                    print("Collision at grid point %d doesn't exist." %
                          grid_points[0])
                vals = _allocate_collision(False,
                                           mesh,
                                           sigma,
                                           sigma_cutoff,
                                           grid_points,
                                           indices,
                                           is_reducible_collision_matrix,
                                           filename)
                if vals:
                    colmat_at_sigma, gamma_at_sigma, temperatures = vals
                else:
                    if log_level:
                        print("Collision at (grid point %d, band index %d) "
                              "doesn't exist." % (grid_points[0], 1))
                    return False

            for i, gp in enumerate(grid_points):
                if not _collect_collision_gp(colmat_at_sigma,
                                             gamma_at_sigma,
                                             temperatures,
                                             mesh,
                                             sigma,
                                             sigma_cutoff,
                                             i,
                                             gp,
                                             indices,
                                             is_reducible_collision_matrix,
                                             filename,
                                             log_level):
                    num_band = colmat_at_sigma.shape[3]
                    for j in range(num_band):
                        if not _collect_collision_band(
                                colmat_at_sigma,
                                gamma_at_sigma,
                                temperatures,
                                mesh,
                                sigma,
                                sigma_cutoff,
                                i,
                                gp,
                                j,
                                indices,
                                is_reducible_collision_matrix,
                                filename,
                                log_level):
                            return False

            if len(sigmas) == 1:
                gamma = gamma_at_sigma
                collision_matrix = colmat_at_sigma
            else:
                gamma.append(gamma_at_sigma[0])
                collision_matrix.append(colmat_at_sigma[0])
            read_from = "grid_points"

    if len(sigmas) > 1:
        temperatures = np.array(temperatures, dtype='double', order='C')
        gamma = np.array(gamma, dtype='double', order='C')
        collision_matrix = np.array(collision_matrix, dtype='double', order='C')

    lbte.set_gamma(gamma)
    lbte.set_collision_matrix(collision_matrix)
    # lbte.set_temperatures invokes allocation of arrays. So this must
    # be called after setting collision_matrix for saving memory
    # space.
    lbte.set_temperatures(temperatures)

    return read_from

def _allocate_collision(for_gps,
                        mesh,
                        sigma,
                        sigma_cutoff,
                        grid_points,
                        indices,
                        is_reducible_collision_matrix,
                        filename):
    num_mesh_points = np.prod(mesh)
    if for_gps:
        collision = read_collision_from_hdf5(mesh,
                                             indices=indices,
                                             grid_point=grid_points[0],
                                             sigma=sigma,
                                             sigma_cutoff=sigma_cutoff,
                                             filename=filename,
                                             verbose=False)
    else:
        collision = read_collision_from_hdf5(mesh,
                                             indices=indices,
                                             grid_point=grid_points[0],
                                             band_index=0,
                                             sigma=sigma,
                                             sigma_cutoff=sigma_cutoff,
                                             filename=filename,
                                             verbose=False)
    if collision is None:
        return False

    num_temp = len(collision[2]) # This is to treat indices="all".
    if is_reducible_collision_matrix:
        if for_gps:
            num_band = collision[0].shape[4] # for gps (s,T,b,irgp,b)
        else:
            num_band = collision[0].shape[3] # for bands (s,T,irgp,b)
        gamma_at_sigma = np.zeros(
            (1, num_temp, num_mesh_points, num_band),
            dtype='double', order='C')
        colmat_at_sigma = np.zeros(
            (1, num_temp,
             num_mesh_points, num_band,
             num_mesh_points, num_band),
            dtype='double', order='C')
    else:
        if for_gps:
            num_band = collision[0].shape[5] # for gps (s,T,b0,3,irgp,b,3)
        else:
            num_band = collision[0].shape[4] # for bands (s,T,3,irgp,b,3)
        gamma_at_sigma = np.zeros(
            (1, num_temp, len(grid_points), num_band),
            dtype='double', order='C')
        colmat_at_sigma = np.zeros(
            (1, num_temp,
             len(grid_points), num_band, 3,
             len(grid_points), num_band, 3),
            dtype='double', order='C')
    temperatures = np.zeros(num_temp, dtype='double', order='C')

    return colmat_at_sigma, gamma_at_sigma, temperatures

def _collect_collision_gp(colmat_at_sigma,
                          gamma_at_sigma,
                          temperatures,
                          mesh,
                          sigma,
                          sigma_cutoff,
                          i,
                          gp,
                          indices,
                          is_reducible_collision_matrix,
                          filename,
                          log_level):
    collision_gp = read_collision_from_hdf5(
        mesh,
        indices=indices,
        grid_point=gp,
        sigma=sigma,
        sigma_cutoff=sigma_cutoff,
        filename=filename,
        verbose=(log_level > 0))
    if log_level:
        sys.stdout.flush()

    if not collision_gp:
        return False

    (colmat_at_gp,
     gamma_at_gp,
     temperatures_at_gp) = collision_gp
    if is_reducible_collision_matrix:
        igp = gp
    else:
        igp = i
    gamma_at_sigma[0, :, igp] = gamma_at_gp
    colmat_at_sigma[0, :, igp] = colmat_at_gp[0]
    temperatures[:] = temperatures_at_gp

    return True

def _collect_collision_band(colmat_at_sigma,
                            gamma_at_sigma,
                            temperatures,
                            mesh,
                            sigma,
                            sigma_cutoff,
                            i,
                            gp,
                            j,
                            indices,
                            is_reducible_collision_matrix,
                            filename,
                            log_level):
    collision_band = read_collision_from_hdf5(
        mesh,
        indices=indices,
        grid_point=gp,
        band_index=j,
        sigma=sigma,
        sigma_cutoff=sigma_cutoff,
        filename=filename,
        verbose=(log_level > 0))
    if log_level:
        sys.stdout.flush()

    if collision_band is False:
        return False

    (colmat_at_band,
     gamma_at_band,
     temperatures_at_band) = collision_band
    if is_reducible_collision_matrix:
        igp = gp
    else:
        igp = i
    gamma_at_sigma[0, :, igp, j] = gamma_at_band
    colmat_at_sigma[0, :, igp, j] = colmat_at_band[0]
    temperatures[:] = temperatures_at_band

    return True

class Conductivity_LBTE(Conductivity):
    def __init__(self,
                 interaction,
                 symmetry,
                 grid_points=None,
                 temperatures=None,
                 sigmas=None,
                 sigma_cutoff=None,
                 is_isotope=False,
                 mass_variances=None,
                 boundary_mfp=None, # in micrometre
                 is_reducible_collision_matrix=False,
                 is_kappa_star=True,
                 gv_delta_q=None, # finite difference for group veolocity
                 is_full_pp=False,
                 pinv_cutoff=1.0e-8,
                 pinv_solver=0,
                 log_level=0):
        self._pp = None
        self._temperatures = None
        self._sigmas = None
        self._sigma_cutoff = None
        self._is_kappa_star = None
        self._gv_delta_q = None
        self._is_full_pp = None
        self._log_level = None
        self._primitive = None
        self._dm = None
        self._frequency_factor_to_THz = None
        self._cutoff_frequency = None
        self._boundary_mfp = None

        self._symmetry = None
        self._point_operations = None
        self._rotations_cartesian = None

        self._grid_points = None
        self._grid_weights = None
        self._grid_address = None
        self._ir_grid_points = None
        self._ir_grid_weights = None

        self._read_gamma = False
        self._read_gamma_iso = False

        self._frequencies = None
        self._cv = None
        self._gv = None
        self._gv_sum2 = None
        self._mfp = None
        self._gamma = None
        self._gamma_iso = None
        self._averaged_pp_interaction = None

        self._mesh = None
        self._conversion_factor = None

        self._is_isotope = None
        self._isotope = None
        self._mass_variances = None
        self._grid_point_count = None

        self._collision_eigenvalues = None

        Conductivity.__init__(self,
                              interaction,
                              symmetry,
                              grid_points=grid_points,
                              temperatures=temperatures,
                              sigmas=sigmas,
                              sigma_cutoff=sigma_cutoff,
                              is_isotope=is_isotope,
                              mass_variances=mass_variances,
                              boundary_mfp=boundary_mfp,
                              is_kappa_star=is_kappa_star,
                              gv_delta_q=gv_delta_q,
                              is_full_pp=is_full_pp,
                              log_level=log_level)

        self._is_reducible_collision_matrix = is_reducible_collision_matrix
        if not self._is_kappa_star:
            self._is_reducible_collision_matrix = True
        self._collision_matrix = None
        self._pinv_cutoff = pinv_cutoff
        self._pinv_solver = pinv_solver

        if grid_points is None:
            self._all_grid_points = True
        else:
            self._all_grid_points = False

        if self._temperatures is not None:
            self._allocate_values()

    def set_kappa_at_sigmas(self):
        if len(self._grid_points) != len(self._ir_grid_points):
            print("Collision matrix is not well created.")
            import sys
            sys.exit(1)
        else:
            self._set_kappa_at_sigmas()

    def set_collision_matrix(self, collision_matrix):
        self._collision_matrix = collision_matrix

    def get_collision_matrix(self):
        return self._collision_matrix

    def get_collision_eigenvalues(self):
        return self._collision_eigenvalues

    def get_mean_free_path(self):
        return self._mfp

    def get_frequencies_all(self):
        return self._frequencies[:np.prod(self._mesh)]

    def get_kappa_RTA(self):
        return self._kappa_RTA

    def get_mode_kappa_RTA(self):
        return self._mode_kappa_RTA

    def _run_at_grid_point(self):
        i = self._grid_point_count
        self._show_log_header(i)
        gp = self._grid_points[i]

        if not self._all_grid_points:
            self._collision_matrix[:] = 0

        if not self._read_gamma:
            self._collision.set_grid_point(gp)

            if self._log_level:
                print("Number of triplets: %d" %
                      len(self._pp.get_triplets_at_q()[0]))

            self._set_collision_matrix_at_sigmas(i)

        if self._is_reducible_collision_matrix:
            igp = gp
        else:
            igp = i
        self._set_harmonic_properties(i, igp)
        if self._isotope is not None:
            gamma_iso = self._get_gamma_isotope_at_sigmas(i)
            band_indices = self._pp.get_band_indices()
            self._gamma_iso[:, igp, :] = gamma_iso[:, band_indices]

        if self._log_level:
            self._show_log(i)

    def _allocate_values(self):
        num_band0 = len(self._pp.get_band_indices())
        num_band = self._primitive.get_number_of_atoms() * 3
        num_ir_grid_points = len(self._ir_grid_points)
        num_temp = len(self._temperatures)
        num_mesh_points = np.prod(self._mesh)

        if self._is_reducible_collision_matrix:
            num_grid_points = num_mesh_points
        else:
            num_grid_points = len(self._grid_points)

        if self._all_grid_points:
            num_stored_grid_points = num_grid_points
        else:
            num_stored_grid_points = 1

        self._kappa = np.zeros((len(self._sigmas), num_temp, 6),
                               dtype='double', order='C')
        self._kappa_RTA = np.zeros((len(self._sigmas), num_temp, 6),
                                   dtype='double', order='C')
        self._gv = np.zeros((num_grid_points, num_band0, 3),
                            dtype='double', order='C')
        self._gv_sum2 = np.zeros((num_grid_points, num_band0, 6),
                                 dtype='double', order='C')
        self._mfp = np.zeros((len(self._sigmas),
                              num_temp,
                              num_grid_points,
                              num_band0,
                              3), dtype='double', order='C')
        self._cv = np.zeros((num_temp, num_grid_points, num_band0),
                            dtype='double', order='C')
        if self._is_full_pp:
            self._averaged_pp_interaction = np.zeros(
                (num_grid_points, num_band0), dtype='double', order='C')

        if self._gamma is None:
            self._gamma = np.zeros((len(self._sigmas),
                                    num_temp,
                                    num_grid_points,
                                    num_band0), dtype='double', order='C')
        if self._isotope is not None:
            self._gamma_iso = np.zeros((len(self._sigmas),
                                        num_grid_points,
                                        num_band0), dtype='double', order='C')

        if self._is_reducible_collision_matrix:
            self._mode_kappa = np.zeros((len(self._sigmas),
                                         num_temp,
                                         num_mesh_points,
                                         num_band,
                                         6), dtype='double', order='C')
            self._mode_kappa_RTA = np.zeros((len(self._sigmas),
                                             num_temp,
                                             num_mesh_points,
                                             num_band,
                                             6), dtype='double', order='C')
            self._collision = CollisionMatrix(
                self._pp,
                is_reducible_collision_matrix=True,
                log_level=self._log_level)
            if self._collision_matrix is None:
                self._collision_matrix = np.empty(
                    (len(self._sigmas), num_temp,
                     num_stored_grid_points, num_band0,
                     num_mesh_points, num_band),
                    dtype='double', order='C')
                self._collision_matrix[:] = 0
            self._collision_eigenvalues = np.zeros(
                (len(self._sigmas), num_temp, num_mesh_points * num_band),
                dtype='double', order='C')
        else:
            self._mode_kappa = np.zeros((len(self._sigmas),
                                         num_temp,
                                         num_grid_points,
                                         num_band0,
                                         6), dtype='double')
            self._mode_kappa_RTA = np.zeros((len(self._sigmas),
                                             num_temp,
                                             num_grid_points,
                                             num_band0,
                                             6), dtype='double')
            self._rot_grid_points = np.zeros(
                (len(self._ir_grid_points), len(self._point_operations)),
                dtype='intc')
            for i, ir_gp in enumerate(self._ir_grid_points):
                self._rot_grid_points[i] = get_grid_points_by_rotations(
                    self._grid_address[ir_gp],
                    self._point_operations,
                    self._mesh)
            self._collision = CollisionMatrix(
                self._pp,
                point_operations=self._point_operations,
                ir_grid_points=self._ir_grid_points,
                rot_grid_points=self._rot_grid_points,
                log_level=self._log_level)
            if self._collision_matrix is None:
                self._collision_matrix = np.empty(
                    (len(self._sigmas),
                     num_temp,
                     num_stored_grid_points, num_band0, 3,
                     num_ir_grid_points, num_band, 3),
                    dtype='double', order='C')
                self._collision_matrix[:] = 0
            self._collision_eigenvalues = np.zeros(
                (len(self._sigmas),
                 num_temp,
                 num_ir_grid_points * num_band * 3),
                dtype='double', order='C')

    def _set_collision_matrix_at_sigmas(self, i):
        for j, sigma in enumerate(self._sigmas):
            if self._log_level:
                text = "Calculating collision matrix with "
                if sigma is None:
                    text += "tetrahedron method"
                else:
                    text += "sigma=%s" % sigma
                    if self._sigma_cutoff is None:
                        text += "."
                    else:
                        text += "(%4.2f SD)." % self._sigma_cutoff
                print(text)

            self._collision.set_sigma(sigma, sigma_cutoff=self._sigma_cutoff)
            self._collision.set_integration_weights()

            if j != 0 and (self._is_full_pp or self._sigma_cutoff is None):
                if self._log_level:
                    print("Existing ph-ph interaction is used.")
            else:
                if self._log_level:
                    print("Calculating ph-ph interaction...")
                self._collision.run_interaction(is_full_pp=self._is_full_pp)
            if self._is_full_pp and j == 0:
                self._averaged_pp_interaction[i] = (
                    self._pp.get_averaged_interaction())

            for k, t in enumerate(self._temperatures):
                self._collision.set_temperature(t)
                self._collision.run()
                if self._all_grid_points:
                    if self._is_reducible_collision_matrix:
                        i_data = self._grid_points[i]
                    else:
                        i_data = i
                else:
                    i_data = 0
                self._gamma[j, k, i_data] = (
                    self._collision.get_imag_self_energy())
                self._collision_matrix[j, k, i_data] = (
                    self._collision.get_collision_matrix())

        self._collision.delete_integration_weights()
        self._pp.delete_interaction_strength()

    def _set_kappa_at_sigmas(self):
        num_band = self._primitive.get_number_of_atoms() * 3
        if self._is_reducible_collision_matrix:
            if self._is_kappa_star:
                self._average_collision_matrix_by_degeneracy()
                self._expand_collisions()
            self._combine_reducible_collisions()
            weights = np.ones(np.prod(self._mesh), dtype='intc')
            self._symmetrize_collision_matrix()
            size = np.prod(self._mesh) * num_band
        else:
            self._combine_collisions()
            weights = self._get_weights()
            for i, w_i in enumerate(weights):
                for j, w_j in enumerate(weights):
                    self._collision_matrix[:, :, i, :, :, j, :, :] *= w_i * w_j
            self._average_collision_matrix_by_degeneracy()
            self._symmetrize_collision_matrix()
            size = len(self._ir_grid_points) * num_band * 3

        for j, sigma in enumerate(self._sigmas):
            if self._log_level:
                text = "----------- Thermal conductivity (W/m-k) "
                if sigma:
                    text += "for sigma=%s -----------" % sigma
                else:
                    text += "with tetrahedron method -----------"
                print(text)
                sys.stdout.flush()

            for k, t in enumerate(self._temperatures):
                if t > 0:
                    self._set_kappa_RTA(j, k, weights)

                    self._set_inv_collision_matrix(j, k, size)
                    self._set_kappa(j, k, weights)
                    # print("spectra")
                    # print(self._get_spectra(j, k, weights))

                    if self._log_level:
                        print(("#%6s       " + " %-10s" * 6) %
                              ("T(K)", "xx", "yy", "zz", "yz", "xz", "xy"))
                        print(("%7.1f " + " %10.3f" * 6) %
                              ((t,) + tuple(self._kappa[j, k])))
                        print((" %6s " + " %10.3f" * 6) %
                              (("(RTA)",) + tuple(self._kappa_RTA[j, k])))
                        print("-" * 76)
                        sys.stdout.flush()

                        sys.stdout.flush()

        if self._log_level:
            print('')

    def _combine_collisions(self):
        num_band = self._primitive.get_number_of_atoms() * 3
        for j, k in list(np.ndindex((len(self._sigmas),
                                     len(self._temperatures)))):
            for i, ir_gp in enumerate(self._ir_grid_points):
                for r, r_gp in zip(self._rotations_cartesian,
                                      self._rot_grid_points[i]):
                    if ir_gp != r_gp:
                        continue

                    main_diagonal = self._get_main_diagonal(i, j, k)
                    for l in range(num_band):
                        self._collision_matrix[
                            j, k, i, l, :, i, l, :] += main_diagonal[l] * r

    def _combine_reducible_collisions(self):
        num_band = self._primitive.get_number_of_atoms() * 3
        num_mesh_points = np.prod(self._mesh)

        for j, k in list(
                np.ndindex((len(self._sigmas), len(self._temperatures)))):
            for i in range(num_mesh_points):
                main_diagonal = self._get_main_diagonal(i, j, k)
                for l in range(num_band):
                    self._collision_matrix[
                        j, k, i, l, i, l] += main_diagonal[l]

    def _expand_collisions(self):
        num_mesh_points = np.prod(self._mesh)
        num_rot = len(self._point_operations)
        num_band = self._primitive.get_number_of_atoms() * 3
        rot_grid_points = np.zeros(
            (num_rot, num_mesh_points), dtype='intc')

        for i in range(num_mesh_points):
            rot_grid_points[:, i] = get_grid_points_by_rotations(
                self._grid_address[i],
                self._point_operations,
                self._mesh)

        for i, ir_gp in enumerate(self._ir_grid_points):
            colmat_irgp = self._collision_matrix[:, :, ir_gp, :, :, :].copy()
            self._collision_matrix[:, :, ir_gp, :, :, :] = 0
            gv_irgp = self._gv[ir_gp].copy()
            self._gv[ir_gp] = 0
            cv_irgp = self._cv[:, ir_gp, :].copy()
            self._cv[:, ir_gp, :] = 0
            gamma_irgp = self._gamma[:, :, ir_gp, :].copy()
            self._gamma[:, :, ir_gp, :] = 0
            multi = (rot_grid_points[:, ir_gp] == ir_gp).sum()
            if self._gamma_iso is not None:
                gamma_iso_irgp = self._gamma_iso[:, ir_gp, :].copy()
                self._gamma_iso[:, ir_gp, :] = 0
            for j, r in enumerate(self._rotations_cartesian):
                gp_r = rot_grid_points[j, ir_gp]
                self._gamma[:, :, gp_r, :] += gamma_irgp / multi
                if self._gamma_iso is not None:
                    self._gamma_iso[:, gp_r, :] += gamma_iso_irgp / multi
                for k in range(num_mesh_points):
                    gp_c = rot_grid_points[j, k]
                    self._collision_matrix[:, :, gp_r, :, gp_c, :] += (
                        colmat_irgp[:, :, :, k, :] / multi)
                self._gv[gp_r] += np.dot(gv_irgp, r.T) / multi
                self._cv[:, gp_r, :] += cv_irgp / multi

    def _get_weights(self):
        """Weights used for collision matrix and |X> and |f>

           r_gps: grid points on k-star with duplicates
                  len(r_gps) == order of crystallographic point group
                  len(unique(r_gps)) == # of arms of the k-star
           self._rot_grid_points contains r_gps for ir-grid points

        """
        weights = []
        for r_gps in self._rot_grid_points:
            weights.append(np.sqrt(len(np.unique(r_gps))) / np.sqrt(len(r_gps)))
        return weights

    def _symmetrize_collision_matrix(self):
        if self._log_level:
            print("- Making collision matrix symmetric...")
            sys.stdout.flush()

        # C-API implementation causes segmentation fault in specific
        # cases. The reason is not clear, but it seems to happen when
        # using large memory space. Passing data to C-API is doubted.
        #
        # import phono3py._phono3py as phono3c
        # print(self._collision_matrix.flags)
        # print(self._collision_matrix.shape)
        # sys.stdout.flush()
        # phono3c.symmetrize_collision_matrix(self._collision_matrix)

        if self._is_reducible_collision_matrix:
            size = np.prod(self._collision_matrix.shape[2:4])
        else:
            size = np.prod(self._collision_matrix.shape[2:5])
        for i in range(self._collision_matrix.shape[0]):
            for j in range(self._collision_matrix.shape[1]):
                col_mat = self._collision_matrix[i, j].reshape(size, size)
                col_mat += col_mat.T
                col_mat /= 2

    def _average_collision_matrix_by_degeneracy(self):
        # Average matrix elements belonging to degenerate bands
        if self._log_level:
            print("- Averaging collision matrix elements "
                  "by phonon degeneracy ...")
            sys.stdout.flush()

        col_mat = self._collision_matrix
        for i, gp in enumerate(self._ir_grid_points):
            freqs = self._frequencies[gp]
            deg_sets = degenerate_sets(freqs)
            for dset in deg_sets:
                bi_set = []
                for j in range(len(freqs)):
                    if j in dset:
                        bi_set.append(j)

                if self._is_reducible_collision_matrix:
                    sum_col = (col_mat[:, :, gp, bi_set, :, :].sum(axis=2) /
                               len(bi_set))
                    for j in bi_set:
                        col_mat[:, :, gp, j, :, :] = sum_col
                else:
                    sum_col = (
                        col_mat[:, :, i, bi_set, :, :, :, :].sum(axis=2) /
                        len(bi_set))
                    for j in bi_set:
                        col_mat[:, :, i, j, :, :, :, :] = sum_col

        for i, gp in enumerate(self._ir_grid_points):
            freqs = self._frequencies[gp]
            deg_sets = degenerate_sets(freqs)
            for dset in deg_sets:
                bi_set = []
                for j in range(len(freqs)):
                    if j in dset:
                        bi_set.append(j)
                if self._is_reducible_collision_matrix:
                    sum_col = (col_mat[:, :, :, :, gp, bi_set].sum(axis=4) /
                               len(bi_set))
                    for j in bi_set:
                        col_mat[:, :, :, :, gp, j] = sum_col
                else:
                    sum_col = (
                        col_mat[:, :, :, :, :, i, bi_set, :].sum(axis=5) /
                        len(bi_set))
                    for j in bi_set:
                        col_mat[:, :, :, :, :, i, j, :] = sum_col

    def _get_X(self, i_temp, weights, gv):
        num_band = self._primitive.get_number_of_atoms() * 3
        X = gv.copy()
        if self._is_reducible_collision_matrix:
            num_mesh_points = np.prod(self._mesh)
            freqs = self._frequencies[:num_mesh_points]
        else:
            freqs = self._frequencies[self._ir_grid_points]

        t = self._temperatures[i_temp]
        sinh = np.where(freqs > self._cutoff_frequency,
                        np.sinh(freqs * THzToEv / (2 * Kb * t)),
                        -1.0)
        inv_sinh = np.where(sinh > 0, 1.0 / sinh, 0)
        freqs_sinh = freqs * THzToEv * inv_sinh / (4 * Kb * t ** 2)

        for i, f in enumerate(freqs_sinh):
            X[i] *= weights[i]
            for j in range(num_band):
                X[i, j] *= f[j]

        if t > 0:
            return X.reshape(-1, 3)
        else:
            return np.zeros_like(X.reshape(-1, 3))

    def _get_Y(self, i_sigma, i_temp, X):
        num_band = self._primitive.get_number_of_atoms() * 3

        if self._is_reducible_collision_matrix:
            num_mesh_points = np.prod(self._mesh)
            size = num_mesh_points * num_band
        else:
            num_ir_grid_points = len(self._ir_grid_points)
            size = num_ir_grid_points * num_band * 3
        v = self._collision_matrix[i_sigma, i_temp].reshape(size, size)

        if self._pinv_solver % 10 in [0, 6]:
            if self._log_level:
                print("Calculating pseudo-inv of collision matrix by np.dot "
                      "(cutoff=%-.1e)" % self._pinv_cutoff)
                sys.stdout.flush()

            w = self._collision_eigenvalues[i_sigma, i_temp]
            e = np.zeros_like(w)
            pinv_method = self._pinv_solver // 10
            if pinv_method not in [0, 1]:
                pinv_method = 0
            for l, val in enumerate(w):
                if (val > self._pinv_cutoff and pinv_method == 1 or
                    abs(val) > self._pinv_cutoff and pinv_method == 0):
                    e[l] = 1 / val
            if self._is_reducible_collision_matrix:
                X1 = np.dot(v.T, X)
                for i in range(3):
                    X1[:, i] *= e
                Y = np.dot(v, X1)
            else:
                Y = np.dot(v, e * np.dot(v.T, X.ravel())).reshape(-1, 3)
        else:
            if self._is_reducible_collision_matrix:
                Y = np.dot(v, X)
            else:
                Y = np.dot(v, X.ravel()).reshape(-1, 3)

        return Y

    def _get_I(self, a, b, size):
        r_sum = np.zeros((3, 3), dtype='double')
        for r in self._rotations_cartesian:
            for i in range(3):
                for j in range(3):
                    r_sum[i, j] += r[a, i] * r[b, j]
        I = np.zeros((size * 3, size * 3), dtype='double')
        for i in range(size):
            I[(i * 3):((i + 1) * 3), (i * 3):((i + 1) * 3)] = r_sum

        return I

    def _set_inv_collision_matrix(self, i_sigma, i_temp, size):
        """Compute pinv of collision matrix
        Args:
           size: The following value is expected:
              ir-colmat:  num_ir_grid_points * num_band * 3
              red-colmat: num_mesh_points * num_band
        """
        solver = self._pinv_solver % 10
        pinv_method = self._pinv_solver // 10
        # Safest choice
        if solver == 9:
            print("*** Attention: This isn't the calculation, but the test. ***")
        elif solver == 0: # default solver
            try:
                import scipy.linalg
            except ImportError:
                solver = 1
            else:
                solver = 6
        elif solver not in range(1, 7):
            solver = 1
        if pinv_method not in [0, 1]:
            pinv_method = 0

        if solver == 1: # dsyev: safer and slower than dsyevd and smallest
                        #        memory usage
            if self._log_level:
                print("Pseudo-inversion (cutoff=%-.1e) by lapacke dsyev..." %
                      self._pinv_cutoff)
                sys.stdout.flush()
            import phono3py._phono3py as phono3c
            w = np.zeros(size, dtype='double')
            phono3c.inverse_collision_matrix(self._collision_matrix,
                                             w,
                                             i_sigma,
                                             i_temp,
                                             self._pinv_cutoff,
                                             solver - 1,
                                             pinv_method)
        elif solver == 2: # dsyevd: faster than dsyev and lagest memory usage
            if self._log_level:
                print("Pseudo-inversion (cutoff=%-.1e) by lapacke dsyevd..." %
                      self._pinv_cutoff)
                sys.stdout.flush()
            import phono3py._phono3py as phono3c
            w = np.zeros(size, dtype='double')
            phono3c.inverse_collision_matrix(self._collision_matrix,
                                             w,
                                             i_sigma,
                                             i_temp,
                                             self._pinv_cutoff,
                                             solver - 1,
                                             pinv_method) # 1 or 11
        elif solver == 3: # np.linalg.eigh depends on dsyevd.
                          # Dangeraous to use this because
                          # the result becomes different.
            if self._log_level:
                print("Diagonalizing by np.linalg.eigh...")
                sys.stdout.flush()
            col_mat = self._collision_matrix[i_sigma, i_temp].reshape(
                size, size)
            w, col_mat[:] = np.linalg.eigh(col_mat)

            if self._log_level:
                print("Calculating pseudo-inv of collision matrix "
                      "(cutoff=%-.1e)" % self._pinv_cutoff)
                sys.stdout.flush()
            import phono3py._phono3py as phono3c
            phono3c.pinv_from_eigensolution(self._collision_matrix,
                                            np.array(w, dtype='double'),
                                            i_sigma,
                                            i_temp,
                                            self._pinv_cutoff,
                                            pinv_method)
        elif solver == 4: # scipy.linalg.lapack.dsyev
            if self._log_level:
                print("Diagonalizing by scipy.linalg.lapack.dsyev...")
                sys.stdout.flush()
            import scipy.linalg
            col_mat = self._collision_matrix[i_sigma, i_temp].reshape(
                size, size)
            w, col_mat[:], info = scipy.linalg.lapack.dsyev(col_mat)

            if self._log_level:
                print("Calculating pseudo-inv of collision matrix "
                      "(cutoff=%-.1e)" % self._pinv_cutoff)
                sys.stdout.flush()
            import phono3py._phono3py as phono3c
            phono3c.pinv_from_eigensolution(self._collision_matrix,
                                            np.array(w, dtype='double'),
                                            i_sigma,
                                            i_temp,
                                            self._pinv_cutoff,
                                            pinv_method)
        elif solver == 5: # scipy.linalg.lapack.dsyevd
                          # This is more stable than numpy.linalg.eigh.
            if self._log_level:
                print("Diagonalizing by scipy.linalg.lapack.dsyevd...")
                sys.stdout.flush()
            import scipy.linalg
            col_mat = self._collision_matrix[i_sigma, i_temp].reshape(
                size, size)
            w, col_mat[:], info = scipy.linalg.lapack.dsyevd(col_mat)
            if self._log_level:
                print("Calculating pseudo-inv of collision matrix "
                      "(cutoff=%-.1e)" % self._pinv_cutoff)
                sys.stdout.flush()
            import phono3py._phono3py as phono3c
            phono3c.pinv_from_eigensolution(self._collision_matrix,
                                            np.array(w, dtype='double'),
                                            i_sigma,
                                            i_temp,
                                            self._pinv_cutoff,
                                            pinv_method)
        elif solver == 6: # fully scipy & numpy
            if self._log_level:
                print("Diagonalizing by scipy.linalg.lapack.dsyev...")
                sys.stdout.flush()
            import scipy.linalg
            col_mat = self._collision_matrix[i_sigma, i_temp].reshape(
                size, size)
            w, col_mat[:], info = scipy.linalg.lapack.dsyev(col_mat)

            # Pseudo inversion is done together with with multiplying
            # X in _set_kappa to save memory space.

        elif solver == 9: # Test
            w = np.ones(size, dtype='double')
            if self._log_level:
                print("Calculating pseudo-inv of collision matrix "
                      "(cutoff=%-.1e)" % self._pinv_cutoff)
                sys.stdout.flush()
            import phono3py._phono3py as phono3c
            phono3c.pinv_from_eigensolution(self._collision_matrix,
                                            np.array(w, dtype='double'),
                                            i_sigma,
                                            i_temp,
                                            self._pinv_cutoff,
                                            pinv_method)

        self._collision_eigenvalues[i_sigma, i_temp] = w

    def _set_kappa(self, i_sigma, i_temp, weights):
        if self._is_reducible_collision_matrix:
            X = self._get_X(i_temp, weights, self._gv)
            num_mesh_points = np.prod(self._mesh)
            Y = self._get_Y(i_sigma, i_temp, X)
            self._set_mean_free_path(i_sigma, i_temp, weights, Y)
            # Putting self._rotations_cartesian is to symmetrize kappa.
            # None can be put instead for watching pure information.
            self._set_mode_kappa(self._mode_kappa,
                                 X,
                                 Y,
                                 num_mesh_points,
                                 self._rotations_cartesian,
                                 i_sigma,
                                 i_temp)
            self._mode_kappa[i_sigma, i_temp] /= len(self._rotations_cartesian)
            self._kappa[i_sigma, i_temp] = (
                self._mode_kappa[i_sigma, i_temp].sum(axis=0).sum(axis=0))
        else:
            X = self._get_X(i_temp, weights, self._gv)
            num_ir_grid_points = len(self._ir_grid_points)
            Y = self._get_Y(i_sigma, i_temp, X)
            self._set_mean_free_path(i_sigma, i_temp, weights, Y)
            self._set_mode_kappa(self._mode_kappa,
                                 X,
                                 Y,
                                 num_ir_grid_points,
                                 self._rotations_cartesian,
                                 i_sigma,
                                 i_temp)
            # self._set_mode_kappa_from_mfp(weights,
            #                               num_ir_grid_points,
            #                               self._rotations_cartesian,
            #                               i_sigma,
            #                               i_temp)

            self._kappa[i_sigma, i_temp] = (
                self._mode_kappa[i_sigma, i_temp].sum(axis=0).sum(axis=0))

    def _set_kappa_RTA(self, i_sigma, i_temp, weights):
        num_band = self._primitive.get_number_of_atoms() * 3
        X = self._get_X(i_temp, weights, self._gv)
        Y = np.zeros_like(X)

        if self._is_reducible_collision_matrix:
            # This RTA is not equivalent to conductivity_RTA.
            # The lifetime is defined from the diagonal part of
            # collision matrix.
            num_mesh_points = np.prod(self._mesh)
            size = num_mesh_points * num_band
            v_diag = np.diagonal(
                self._collision_matrix[i_sigma, i_temp].reshape(size, size))

            for gp in range(num_mesh_points):
                frequencies = self._frequencies[gp]
                for j, f in enumerate(frequencies):
                    if f > self._cutoff_frequency:
                        i_mode = gp * num_band + j
                        Y[i_mode, :] = X[i_mode, :] / v_diag[i_mode]
            # Putting self._rotations_cartesian is to symmetrize kappa.
            # None can be put instead for watching pure information.
            self._set_mode_kappa(self._mode_kappa_RTA,
                                 X,
                                 Y,
                                 num_mesh_points,
                                 self._rotations_cartesian,
                                 i_sigma,
                                 i_temp)
            g = len(self._rotations_cartesian)
            self._mode_kappa_RTA[i_sigma, i_temp] /= g
            self._kappa_RTA[i_sigma, i_temp] = (
                self._mode_kappa_RTA[i_sigma, i_temp].sum(axis=0).sum(axis=0))
        else:
            # This RTA is supposed to be the same as conductivity_RTA.
            num_ir_grid_points = len(self._ir_grid_points)
            size = num_ir_grid_points * num_band * 3
            for i, gp in enumerate(self._ir_grid_points):
                g = self._get_main_diagonal(i, i_sigma, i_temp)
                frequencies = self._frequencies[gp]
                for j, f in enumerate(frequencies):
                    if f > self._cutoff_frequency:
                        i_mode = i * num_band + j
                        old_settings = np.seterr(all='raise')
                        try:
                            Y[i_mode, :] = X[i_mode, :] / g[j]
                        except:
                            print("=" * 26 + " Warning " + "=" * 26)
                            print(" Unexpected physical condition of ph-ph "
                                  "interaction calculation was found.")
                            print(" g[j]=%f at gp=%d, band=%d, freq=%f" %
                                  (g[j], gp, j + 1, f))
                            print("=" * 61)
                        np.seterr(**old_settings)

            self._set_mode_kappa(self._mode_kappa_RTA,
                                 X,
                                 Y,
                                 num_ir_grid_points,
                                 self._rotations_cartesian,
                                 i_sigma,
                                 i_temp)
            self._kappa_RTA[i_sigma, i_temp] = (
                self._mode_kappa_RTA[i_sigma, i_temp].sum(axis=0).sum(axis=0))

    def _set_mode_kappa(self,
                        mode_kappa,
                        X,
                        Y,
                        num_grid_points,
                        rotations_cartesian,
                        i_sigma,
                        i_temp):
        num_band = self._primitive.get_number_of_atoms() * 3
        for i, (v_gp, f_gp) in enumerate(zip(X.reshape(num_grid_points,
                                                       num_band, 3),
                                             Y.reshape(num_grid_points,
                                                       num_band, 3))):
            for j, (v, f) in enumerate(zip(v_gp, f_gp)):
                # Do not consider three lowest modes at Gamma-point
                # It is assumed that there are no imaginary modes.
                if (self._grid_address[i] == 0).all() and j < 3:
                    continue

                if rotations_cartesian is None:
                    sum_k = np.outer(v, f)
                else:
                    sum_k = np.zeros((3, 3), dtype='double')
                    for r in rotations_cartesian:
                        sum_k += np.outer(np.dot(r, v), np.dot(r, f))
                sum_k = sum_k + sum_k.T
                for k, vxf in enumerate(
                        ((0, 0), (1, 1), (2, 2), (1, 2), (0, 2), (0, 1))):
                    mode_kappa[i_sigma, i_temp, i, j, k] = sum_k[vxf]

        t = self._temperatures[i_temp]
        # Collision matrix is half of that defined in Chaput's paper.
        # Therefore here 2 is not necessary multiplied.
        # sum_k = sum_k + sum_k.T is equivalent to I(a,b) + I(b,a).
        mode_kappa[i_sigma, i_temp] *= (
            self._conversion_factor * Kb * t ** 2 / np.prod(self._mesh))

    def _set_mode_kappa_from_mfp(self,
                                 weights,
                                 num_grid_points,
                                 rotations_cartesian,
                                 i_sigma,
                                 i_temp):
        for i, (v_gp, mfp_gp, cv_gp) in enumerate(
                zip(self._gv, self._mfp[i_sigma, i_temp], self._cv[i_temp])):
            for j, (v, mfp, cv) in enumerate(zip(v_gp, mfp_gp, cv_gp)):
                sum_k = np.zeros((3, 3), dtype='double')
                for r in rotations_cartesian:
                    sum_k += np.outer(np.dot(r, v), np.dot(r, mfp))
                sum_k = (sum_k + sum_k.T) / 2 * cv * weights[i] ** 2 * 2 * np.pi
                for k, vxf in enumerate(
                        ((0, 0), (1, 1), (2, 2), (1, 2), (0, 2), (0, 1))):
                    self._mode_kappa[i_sigma, i_temp, i, j, k] = sum_k[vxf]
        self._mode_kappa *= - self._conversion_factor / np.prod(self._mesh)

    def _set_mean_free_path(self, i_sigma, i_temp, weights, Y):
        if self._is_reducible_collision_matrix:
            num_grid_points = np.prod(self._mesh)
        else:
            num_grid_points = len(self._ir_grid_points)
        t = self._temperatures[i_temp]
        num_band = self._primitive.get_number_of_atoms() * 3
        # Collision matrix is half of that defined in Chaput's paper.
        # Therefore Y is divided by 2.
        for i, f_gp in enumerate((Y / 2).reshape(num_grid_points, num_band, 3)):
            for j, f in enumerate(f_gp):
                cv = self._cv[i_temp, i, j]
                if cv < 1e-10:
                    continue
                self._mfp[i_sigma, i_temp, i, j] = (
                    - 2 * t * np.sqrt(Kb / cv) / weights[i] * f / (2 * np.pi))

    def _get_spectra(self, i_sigma, i_temp, weights):
        import scipy.linalg
        num_band = self._primitive.get_number_of_atoms() * 3
        num_ir_grid_points = len(self._ir_grid_points)
        inv_col_mat = self._collision_matrix[i_sigma, i_temp].reshape(
            num_ir_grid_points * num_band * 3,
            num_ir_grid_points * num_band * 3)
        X = self._get_X(i_temp, weights, self._gv)
        num_band = self._primitive.get_number_of_atoms() * 3
        I = self._get_I(2, 2, num_ir_grid_points * num_band)
        inv_col_mat[:] = np.dot(inv_col_mat, I.T)
        w, inv_col_mat[:], info = scipy.linalg.lapack.dsyev(inv_col_mat)
        for i, x in enumerate(w):
            if i % 8 == 0:
                print("")
            sys.stdout.write("%f " % x)

    def _show_log(self, i):
        q = self._qpoints[i]
        gp = self._grid_points[i]
        frequencies = self._frequencies[gp]
        if self._is_reducible_collision_matrix:
            gv = self._gv[gp]
        else:
            gv = self._gv[i]
        if self._is_full_pp:
            ave_pp = self._averaged_pp_interaction[i]
            text = "Frequency     group velocity (x, y, z)     |gv|       Pqj"
        else:
            text = "Frequency     group velocity (x, y, z)     |gv|"

        if self._gv_delta_q is None:
            pass
        else:
            text += "  (dq=%3.1e)" % self._gv_delta_q
        print(text)
        if self._is_full_pp:
            for f, v, pp in zip(frequencies, gv, ave_pp):
                print("%8.3f   (%8.3f %8.3f %8.3f) %8.3f %11.3e" %
                      (f, v[0], v[1], v[2], np.linalg.norm(v), pp))
        else:
            for f, v in zip(frequencies, gv):
                print("%8.3f   (%8.3f %8.3f %8.3f) %8.3f" %
                      (f, v[0], v[1], v[2], np.linalg.norm(v)))

        sys.stdout.flush()

    def _py_symmetrize_collision_matrix(self):
        num_band = self._primitive.get_number_of_atoms() * 3
        num_ir_grid_points = len(self._ir_grid_points)
        for i in range(num_ir_grid_points):
            for j in range(num_band):
                for k in range(3):
                    for l in range(num_ir_grid_points):
                        for m in range(num_band):
                            for n in range(3):
                                self._py_set_symmetrized_element(
                                    i, j, k, l, m, n)

    def _py_set_symmetrized_element(self, i, j, k, l, m, n):
        sym_val = (self._collision_matrix[:, :, i, j, k, l, m, n] +
                   self._collision_matrix[:, :, l, m, n, i, j, k]) / 2
        self._collision_matrix[:, :, i, j, k, l, m, n] = sym_val
        self._collision_matrix[:, :, l, m, n, i, j, k] = sym_val

    def _py_symmetrize_collision_matrix_no_kappa_stars(self):
        num_band = self._primitive.get_number_of_atoms() * 3
        num_ir_grid_points = len(self._ir_grid_points)
        for i in range(num_ir_grid_points):
            for j in range(num_band):
                for k in range(num_ir_grid_points):
                    for l in range(num_band):
                        self._py_set_symmetrized_element_no_kappa_stars(
                            i, j, k, l)

    def _py_set_symmetrized_element_no_kappa_stars(self, i, j, k, l):
        sym_val = (self._collision_matrix[:, :, i, j, k, l] +
                   self._collision_matrix[:, :, k, l, i, j]) / 2
        self._collision_matrix[:, :, i, j, k, l] = sym_val
        self._collision_matrix[:, :, k, l, i, j] = sym_val



# Copyright (C) 2015 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys
import numpy as np
from phono3py.file_IO import (write_ir_grid_points,
                              write_grid_address_to_hdf5)
from phono3py.phonon3.triplets import (get_coarse_ir_grid_points,
                                       get_number_of_triplets)

def write_grid_points(primitive,
                      mesh,
                      mesh_divs,
                      coarse_mesh_shifts,
                      is_kappa_star,
                      symprec,
                      log_level):
    print("-" * 76)
    if mesh is None:
        print("To write grid points, mesh numbers have to be specified.")
    else:
        (ir_grid_points,
         grid_weights,
         bz_grid_address,
         grid_mapping_table) = get_coarse_ir_grid_points(
             primitive,
             mesh,
             mesh_divs,
             coarse_mesh_shifts,
             is_kappa_star=is_kappa_star,
             symprec=symprec)
        write_ir_grid_points(mesh,
                             mesh_divs,
                             ir_grid_points,
                             grid_weights,
                             bz_grid_address,
                             np.linalg.inv(primitive.get_cell()))
        gadrs_hdf5_fname = write_grid_address_to_hdf5(bz_grid_address,
                                                      mesh,
                                                      grid_mapping_table)

        print("Ir-grid points are written into \"ir_grid_points.yaml\".")
        print("Grid addresses are written into \"%s\"." % gadrs_hdf5_fname)

def show_num_triplets(primitive,
                      mesh,
                      mesh_divs,
                      grid_points,
                      coarse_mesh_shifts,
                      is_kappa_star,
                      symprec,
                      log_level):
    print("-" * 76)

    ir_grid_points, _, grid_address, _ = get_coarse_ir_grid_points(
        primitive,
        mesh,
        mesh_divs,
        coarse_mesh_shifts,
        is_kappa_star=is_kappa_star,
        symprec=symprec)

    if grid_points:
        _grid_points = grid_points
    else:
        _grid_points = ir_grid_points

    print("Grid point        q-point        No. of triplets")
    for gp in _grid_points:
        num_triplets = get_number_of_triplets(primitive,
                                              mesh,
                                              gp,
                                              symprec=symprec)
        q = grid_address[gp] / np.array(mesh, dtype='double')
        print("  %5d     (%5.2f %5.2f %5.2f)  %8d" %
              (gp, q[0], q[1], q[2], num_triplets))



# Copyright (C) 2015 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from phonopy.units import VaspToTHz
from phono3py.phonon3.triplets import get_grid_point_from_address

def get_phono3py_configurations(settings):
    primitive_matrix = settings.get_primitive_matrix()
    supercell_matrix = settings.get_supercell_matrix()
    phonon_supercell_matrix = settings.get_phonon_supercell_matrix()
    masses = settings.get_masses()
    mesh = settings.get_mesh_numbers()
    mesh_divs = settings.get_mesh_divisors()
    coarse_mesh_shifts = settings.get_coarse_mesh_shifts()
    grid_points = settings.get_grid_points()
    grid_addresses = settings.get_grid_addresses()
    if grid_addresses is not None:
        grid_points = [get_grid_point_from_address(ga, mesh)
                       for ga in grid_addresses]
    band_indices = settings.get_band_indices()

    # Brillouin zone integration: Tetrahedron (default) or smearing method
    sigma = settings.get_sigma()
    if sigma is None:
        sigmas = []
    elif isinstance(sigma, float):
        sigmas = [sigma]
    else:
        sigmas = sigma
    if settings.get_is_tetrahedron_method():
        sigmas = [None] + sigmas
    if len(sigmas) == 0:
        sigmas = [None]

    sigma_cutoff = settings.get_sigma_cutoff_width()

    if settings.get_temperatures() is None:
        if settings.get_is_joint_dos():
            temperature_points = None
            temperatures = None
        else:
            t_max = settings.get_max_temperature()
            t_min = settings.get_min_temperature()
            t_step = settings.get_temperature_step()
            temperature_points = [0.0, 300.0] # For spectra
            temperatures = np.arange(t_min, t_max + float(t_step) / 10, t_step)
    else:
        temperature_points = settings.get_temperatures() # For spectra
        temperatures = settings.get_temperatures() # For others

    if settings.get_frequency_conversion_factor() is None:
        frequency_factor_to_THz = VaspToTHz
    else:
        frequency_factor_to_THz = settings.get_frequency_conversion_factor()

    if settings.get_num_frequency_points() is None:
        if settings.get_frequency_pitch() is None:
            num_frequency_points = 201
            frequency_step = None
        else:
            num_frequency_points = None
            frequency_step = settings.get_frequency_pitch()
    else:
        num_frequency_points = settings.get_num_frequency_points()
        frequency_step = None

    if settings.get_frequency_scale_factor() is None:
        frequency_scale_factor = 1.0
    else:
        frequency_scale_factor = settings.get_frequency_scale_factor()

    if settings.get_cutoff_frequency() is None:
        cutoff_frequency = 1e-2
    else:
        cutoff_frequency = settings.get_cutoff_frequency()

    conf = {}
    conf['primitive_matrix'] = primitive_matrix
    conf['supercell_matrix'] = supercell_matrix
    conf['phonon_supercell_matrix'] = phonon_supercell_matrix
    conf['masses'] = masses
    conf['mesh'] = mesh
    conf['mesh_divs'] = mesh_divs
    conf['coarse_mesh_shifts'] = coarse_mesh_shifts
    conf['grid_points'] = grid_points
    conf['band_indices'] = band_indices
    conf['sigmas'] = sigmas
    conf['sigma_cutoff'] = sigma_cutoff
    conf['temperature_points'] = temperature_points
    conf['temperatures'] = temperatures
    conf['frequency_factor_to_THz'] = frequency_factor_to_THz
    conf['num_frequency_points'] = num_frequency_points
    conf['frequency_step'] = frequency_step
    conf['frequency_scale_factor'] = frequency_scale_factor
    conf['cutoff_frequency'] = cutoff_frequency

    return conf



import numpy as np
from phonopy.cui.settings import Settings, ConfParser, fracval

class Phono3pySettings(Settings):
    def __init__(self):
        Settings.__init__(self)

        self._boundary_mfp = 1.0e6 # In micrometre. The default value is
                                   # just set to avoid divergence.
        self._coarse_mesh_shifts = None
        self._constant_averaged_pp_interaction = None
        self._create_displacements = False
        self._cutoff_fc3_distance = None
        self._cutoff_pair_distance = None
        self._frequency_scale_factor = None
        self._gamma_conversion_factor = None
        self._grid_addresses = None
        self._grid_points = None
        self._ion_clamped = False
        self._is_bterta = False
        self._is_frequency_shift = False
        self._is_full_pp = False
        self._is_gruneisen = False
        self._is_imag_self_energy = False
        self._is_isotope = False
        self._is_joint_dos = False
        self._is_kappa_star = True
        self._is_lbte = False
        self._is_linewidth = False
        self._is_N_U = False
        self._is_reducible_collision_matrix = False
        self._is_symmetrize_fc2 = False
        self._is_symmetrize_fc3_q = False
        self._is_symmetrize_fc3_r = False
        self._mass_variances = None
        self._max_freepath = None
        self._mesh_divisors = None
        self._read_amplitude = False
        self._read_collision = None
        self._read_fc2 = False
        self._read_fc3 = False
        self._read_gamma = False
        self._read_phonon = False
        self._phonon_supercell_matrix = None
        self._pinv_cutoff = 1.0e-8
        self._pinv_solver = 0
        self._pp_conversion_factor = None
        self._scattering_event_class = None # scattering event class 1 or 2
        self._sigma_cutoff_width = None
        self._temperatures = None
        self._use_alm = False
        self._use_ave_pp = False
        self._write_amplitude = False
        self._write_collision = False
        self._write_gamma_detail = False
        self._write_gamma = False
        self._write_phonon = False

    def set_boundary_mfp(self, boundary_mfp):
        self._boundary_mfp = boundary_mfp

    def get_boundary_mfp(self):
        return self._boundary_mfp

    def set_coarse_mesh_shifts(self, coarse_mesh_shifts):
        self._coarse_mesh_shifts = coarse_mesh_shifts

    def get_coarse_mesh_shifts(self):
        return self._coarse_mesh_shifts

    def set_create_displacements(self, create_displacements):
        self._create_displacements = create_displacements

    def get_create_displacements(self):
        return self._create_displacements

    def set_constant_averaged_pp_interaction(self, ave_pp):
        self._constant_averaged_pp_interaction = ave_pp

    def get_constant_averaged_pp_interaction(self):
        return self._constant_averaged_pp_interaction

    def set_cutoff_fc3_distance(self, cutoff_fc3_distance):
        self._cutoff_fc3_distance = cutoff_fc3_distance

    def get_cutoff_fc3_distance(self):
        return self._cutoff_fc3_distance

    def set_cutoff_pair_distance(self, cutoff_pair_distance):
        self._cutoff_pair_distance = cutoff_pair_distance

    def get_cutoff_pair_distance(self):
        return self._cutoff_pair_distance

    def set_frequency_scale_factor(self, frequency_scale_factor):
        self._frequency_scale_factor = frequency_scale_factor

    def get_frequency_scale_factor(self):
        return self._frequency_scale_factor

    def set_gamma_conversion_factor(self, gamma_conversion_factor):
        self._gamma_conversion_factor = gamma_conversion_factor

    def get_gamma_conversion_factor(self):
        return self._gamma_conversion_factor

    def set_grid_addresses(self, grid_addresses):
        self._grid_addresses = grid_addresses

    def get_grid_addresses(self):
        return self._grid_addresses

    def set_grid_points(self, grid_points):
        self._grid_points = grid_points

    def get_grid_points(self):
        return self._grid_points

    def set_ion_clamped(self, ion_clamped):
        self._ion_clamped = ion_clamped

    def get_ion_clamped(self):
        return self._ion_clamped

    def set_is_bterta(self, is_bterta):
        self._is_bterta = is_bterta

    def get_is_bterta(self):
        return self._is_bterta

    def set_is_frequency_shift(self, is_frequency_shift):
        self._is_frequency_shift = is_frequency_shift

    def get_is_frequency_shift(self):
        return self._is_frequency_shift

    def set_is_full_pp(self, is_full_pp):
        self._is_full_pp = is_full_pp

    def get_is_full_pp(self):
        return self._is_full_pp

    def set_is_gruneisen(self, is_gruneisen):
        self._is_gruneisen = is_gruneisen

    def get_is_gruneisen(self):
        return self._is_gruneisen

    def set_is_imag_self_energy(self, is_imag_self_energy):
        self._is_imag_self_energy = is_imag_self_energy

    def get_is_imag_self_energy(self):
        return self._is_imag_self_energy

    def set_is_isotope(self, is_isotope):
        self._is_isotope = is_isotope

    def get_is_isotope(self):
        return self._is_isotope

    def set_is_joint_dos(self, is_joint_dos):
        self._is_joint_dos = is_joint_dos

    def get_is_joint_dos(self):
        return self._is_joint_dos

    def set_is_kappa_star(self, is_kappa_star):
        self._is_kappa_star = is_kappa_star

    def get_is_kappa_star(self):
        return self._is_kappa_star

    def set_is_lbte(self, is_lbte):
        self._is_lbte = is_lbte

    def get_is_lbte(self):
        return self._is_lbte

    def set_is_linewidth(self, is_linewidth):
        self._is_linewidth = is_linewidth

    def get_is_linewidth(self):
        return self._is_linewidth

    def set_is_N_U(self, is_N_U):
        self._is_N_U = is_N_U

    def get_is_N_U(self):
        return self._is_N_U

    def set_is_reducible_collision_matrix(self, is_reducible_collision_matrix):
        self._is_reducible_collision_matrix = is_reducible_collision_matrix

    def get_is_reducible_collision_matrix(self):
        return self._is_reducible_collision_matrix

    def set_is_symmetrize_fc2(self, is_symmetrize_fc2):
        self._is_symmetrize_fc2 = is_symmetrize_fc2

    def get_is_symmetrize_fc2(self):
        return self._is_symmetrize_fc2

    def set_is_symmetrize_fc3_q(self, is_symmetrize_fc3_q):
        self._is_symmetrize_fc3_q = is_symmetrize_fc3_q

    def get_is_symmetrize_fc3_q(self):
        return self._is_symmetrize_fc3_q

    def set_is_symmetrize_fc3_r(self, is_symmetrize_fc3_r):
        self._is_symmetrize_fc3_r = is_symmetrize_fc3_r

    def get_is_symmetrize_fc3_r(self):
        return self._is_symmetrize_fc3_r

    def set_mass_variances(self, mass_variances):
        self._mass_variances = mass_variances

    def get_mass_variances(self):
        return self._mass_variances

    def set_max_freepath(self, max_freepath):
        self._max_freepath = max_freepath

    def get_max_freepath(self):
        return self._max_freepath

    def set_mesh_divisors(self, mesh_divisors):
        self._mesh_divisors = mesh_divisors

    def get_mesh_divisors(self):
        return self._mesh_divisors

    def set_phonon_supercell_matrix(self, matrix):
        self._phonon_supercell_matrix = matrix

    def get_phonon_supercell_matrix(self):
        return self._phonon_supercell_matrix

    def set_pinv_cutoff(self, pinv_cutoff):
        self._pinv_cutoff = pinv_cutoff

    def get_pinv_cutoff(self):
        return self._pinv_cutoff

    def set_pinv_solver(self, pinv_solver):
        self._pinv_solver = pinv_solver

    def get_pinv_solver(self):
        return self._pinv_solver

    def set_pp_conversion_factor(self, pp_conversion_factor):
        self._pp_conversion_factor = pp_conversion_factor

    def get_pp_conversion_factor(self):
        return self._pp_conversion_factor

    def set_read_amplitude(self, read_amplitude):
        self._read_amplitude = read_amplitude

    def get_read_amplitude(self):
        return self._read_amplitude

    def set_read_collision(self, read_collision):
        self._read_collision = read_collision

    def get_read_collision(self):
        return self._read_collision

    def set_read_fc2(self, read_fc2):
        self._read_fc2 = read_fc2

    def get_read_fc2(self):
        return self._read_fc2

    def set_read_fc3(self, read_fc3):
        self._read_fc3 = read_fc3

    def get_read_fc3(self):
        return self._read_fc3

    def set_read_gamma(self, read_gamma):
        self._read_gamma = read_gamma

    def get_read_gamma(self):
        return self._read_gamma

    def set_read_phonon(self, read_phonon):
        self._read_phonon = read_phonon

    def get_read_phonon(self):
        return self._read_phonon

    def set_scattering_event_class(self, scattering_event_class):
        self._scattering_event_class = scattering_event_class

    def get_scattering_event_class(self):
        return self._scattering_event_class

    def set_sigma_cutoff_width(self, sigma_cutoff_width):
        self._sigma_cutoff_width = sigma_cutoff_width

    def get_sigma_cutoff_width(self):
        return self._sigma_cutoff_width

    def set_temperatures(self, temperatures):
        self._temperatures = temperatures

    def get_temperatures(self):
        return self._temperatures

    def set_use_alm(self, use_alm):
        self._use_alm = use_alm

    def get_use_alm(self):
        return self._use_alm

    def set_use_ave_pp(self, use_ave_pp):
        self._use_ave_pp = use_ave_pp

    def get_use_ave_pp(self):
        return self._use_ave_pp

    def set_write_amplitude(self, write_amplitude):
        self._write_amplitude = write_amplitude

    def get_write_amplitude(self):
        return self._write_amplitude

    def set_write_collision(self, write_collision):
        self._write_collision = write_collision

    def get_write_collision(self):
        return self._write_collision

    def set_write_gamma_detail(self, write_gamma_detail):
        self._write_gamma_detail = write_gamma_detail

    def get_write_gamma_detail(self):
        return self._write_gamma_detail

    def set_write_gamma(self, write_gamma):
        self._write_gamma = write_gamma

    def get_write_gamma(self):
        return self._write_gamma

    def set_write_phonon(self, write_phonon):
        self._write_phonon = write_phonon

    def get_write_phonon(self):
        return self._write_phonon

class Phono3pyConfParser(ConfParser):
    def __init__(self, filename=None, args=None):
        ConfParser.__init__(self, filename, args)
        self._read_options()
        self._parse_conf()
        self._settings = Phono3pySettings()
        self._set_settings()

    def _read_options(self):
        if 'phonon_supercell_dimension' in self._args:
            dim_fc2 = self._args.phonon_supercell_dimension
            if dim_fc2 is not None:
                self._confs['dim_fc2'] = dim_fc2

        if 'boundary_mfp' in self._args:
            if self._args.boundary_mfp is not None:
                self._confs['boundary_mfp'] = self._args.boundary_mfp

        if 'constant_averaged_pp_interaction' in self._args:
            const_ave_pp = self._args.constant_averaged_pp_interaction
            if const_ave_pp is not None:
                self._confs['constant_averaged_pp_interaction'] = const_ave_pp

        if 'cutoff_fc3_distance' in self._args:
            cutoff_fc3 = self._args.cutoff_fc3_distance
            if cutoff_fc3 is not None:
                self._confs['cutoff_fc3_distance'] = cutoff_fc3

        if 'cutoff_pair_distance' in self._args:
            cutoff_pair = self._args.cutoff_pair_distance
            if cutoff_pair is not None:
                self._confs['cutoff_pair_distance'] = cutoff_pair

        if 'frequency_scale_factor' in self._args:
            freq_scale = self._args.frequency_scale_factor
            if freq_scale is not None:
                self._confs['frequency_scale_factor'] = freq_scale

        if 'gamma_conversion_factor' in self._args:
            g_conv_factor = self._args.gamma_conversion_factor
            if g_conv_factor is not None:
                self._confs['gamma_conversion_factor'] = g_conv_factor

        if 'grid_addresses' in self._args:
            if self._args.grid_addresses is not None:
                self._confs['grid_addresses'] = self._args.grid_addresses

        if 'grid_points' in self._args:
            if self._args.grid_points is not None:
                self._confs['grid_points'] = self._args.grid_points

        if 'ion_clamped' in self._args:
            if self._args.ion_clamped:
                self._confs['ion_clamped'] = '.true.'

        if 'is_bterta' in self._args:
            if self._args.is_bterta:
                self._confs['bterta'] = '.true.'

        if 'is_gruneisen' in self._args:
            if self._args.is_gruneisen:
                self._confs['gruneisen'] = '.true.'

        if 'is_displacement' in self._args:
            if self._args.is_displacement:
                self._confs['create_displacements'] = '.true.'

        if 'is_frequency_shift' in self._args:
            if self._args.is_frequency_shift:
                self._confs['frequency_shift'] = '.true.'

        if 'is_full_pp' in self._args:
            if self._args.is_full_pp:
                self._confs['full_pp'] = '.true.'

        if 'is_imag_self_energy' in self._args:
            if self._args.is_imag_self_energy:
                self._confs['imag_self_energy'] = '.true.'

        if 'is_isotope' in self._args:
            if self._args.is_isotope:
                self._confs['isotope'] = '.true.'

        if 'is_joint_dos' in self._args:
            if self._args.is_joint_dos:
                self._confs['joint_dos'] = '.true.'

        if 'no_kappa_stars' in self._args:
            if self._args.no_kappa_stars:
                self._confs['kappa_star'] = '.false.'

        if 'is_lbte' in self._args:
            if self._args.is_lbte:
                self._confs['lbte'] = '.true.'

        if 'is_linewidth' in self._args:
            if self._args.is_linewidth:
                self._confs['linewidth'] = '.true.'

        if 'is_N_U' in self._args:
            if self._args.is_N_U:
                self._confs['N_U'] = '.true.'

        if 'is_reducible_collision_matrix' in self._args:
            if self._args.is_reducible_collision_matrix:
                self._confs['reducible_collision_matrix'] = '.true.'

        if 'is_symmetrize_fc2' in self._args:
            if self._args.is_symmetrize_fc2:
                self._confs['symmetrize_fc2'] = '.true.'

        if 'is_symmetrize_fc3_q' in self._args:
            if self._args.is_symmetrize_fc3_q:
                self._confs['symmetrize_fc3_q'] = '.true.'

        if 'is_symmetrize_fc3_r' in self._args:
            if self._args.is_symmetrize_fc3_r:
                self._confs['symmetrize_fc3_r'] = '.true.'

        if 'mass_variances' in self._args:
            if self._args.mass_variances is not None:
                self._confs['mass_variances'] = self._args.mass_variances

        if 'max_freepath' in self._args:
            if self._args.max_freepath is not None:
                self._confs['max_freepath'] = self._args.max_freepath

        if 'mesh_divisors' in self._args:
            if self._args.mesh_divisors is not None:
                self._confs['mesh_divisors'] = self._args.mesh_divisors

        if 'pinv_cutoff' in self._args:
            if self._args.pinv_cutoff is not None:
                self._confs['pinv_cutoff'] = self._args.pinv_cutoff

        if 'pinv_solver' in self._args:
            if self._args.pinv_solver is not None:
                self._confs['pinv_solver'] = self._args.pinv_solver

        if 'pp_conversion_factor' in self._args:
            pp_conv_factor = self._args.pp_conversion_factor
            if pp_conv_factor is not None:
                self._confs['pp_conversion_factor'] = pp_conv_factor

        if 'read_amplitude' in self._args:
            if self._args.read_amplitude:
                self._confs['read_amplitude'] = '.true.'

        if 'read_fc2' in self._args:
            if self._args.read_fc2:
                self._confs['read_fc2'] = '.true.'

        if 'read_fc3' in self._args:
            if self._args.read_fc3:
                self._confs['read_fc3'] = '.true.'

        if 'read_gamma' in self._args:
            if self._args.read_gamma:
                self._confs['read_gamma'] = '.true.'

        if 'read_phonon' in self._args:
            if self._args.read_phonon:
                self._confs['read_phonon'] = '.true.'

        if 'read_collision' in self._args:
            if self._args.read_collision is not None:
                self._confs['read_collision'] = self._args.read_collision

        if 'scattering_event_class' in self._args:
            scatt_class = self._args.scattering_event_class
            if scatt_class is not None:
                self._confs['scattering_event_class'] = scatt_class

        if 'sigma_cutoff_width' in self._args:
            sigma_cutoff = self._args.sigma_cutoff_width
            if sigma_cutoff is not None:
                self._confs['sigma_cutoff_width'] = sigma_cutoff

        if 'temperatures' in self._args:
            if self._args.temperatures is not None:
                self._confs['temperatures'] = self._args.temperatures

        if 'use_alm' in self._args:
            if self._args.use_alm:
                self._confs['use_alm'] = '.true.'

        if 'use_ave_pp' in self._args:
            if self._args.use_ave_pp:
                self._confs['use_ave_pp'] = '.true.'

        if 'write_amplitude' in self._args:
            if self._args.write_amplitude:
                self._confs['write_amplitude'] = '.true.'

        if 'write_gamma_detail' in self._args:
            if self._args.write_gamma_detail:
                self._confs['write_gamma_detail'] = '.true.'

        if 'write_gamma' in self._args:
            if self._args.write_gamma:
                self._confs['write_gamma'] = '.true.'

        if 'write_collision' in self._args:
            if self._args.write_collision:
                self._confs['write_collision'] = '.true.'

        if 'write_phonon' in self._args:
            if self._args.write_phonon:
                self._confs['write_phonon'] = '.true.'

    def _parse_conf(self):
        confs = self._confs

        for conf_key in confs.keys():
            if conf_key == 'create_displacements':
                if confs['create_displacements'].lower() == '.true.':
                    self.set_parameter('create_displacements', True)

            if conf_key == 'dim_fc2':
                matrix = [ int(x) for x in confs['dim_fc2'].split() ]
                if len(matrix) == 9:
                    matrix = np.array(matrix).reshape(3, 3)
                elif len(matrix) == 3:
                    matrix = np.diag(matrix)
                else:
                    self.setting_error(
                        "Number of elements of dim2 has to be 3 or 9.")

                if matrix.shape == (3, 3):
                    if np.linalg.det(matrix) < 1:
                        self.setting_error(
                            "Determinant of supercell matrix has " +
                            "to be positive.")
                    else:
                        self.set_parameter('dim_fc2', matrix)

            if conf_key == 'boundary_mfp':
                self.set_parameter('boundary_mfp',
                                   float(confs['boundary_mfp']))

            if conf_key == 'constant_averaged_pp_interaction':
                self.set_parameter(
                    'constant_averaged_pp_interaction',
                    float(confs['constant_averaged_pp_interaction']))

            if conf_key == 'cutoff_fc3_distance':
                self.set_parameter('cutoff_fc3_distance',
                                   float(confs['cutoff_fc3_distance']))

            if conf_key == 'cutoff_pair_distance':
                self.set_parameter('cutoff_pair_distance',
                                   float(confs['cutoff_pair_distance']))

            if conf_key == 'frequency_scale_factor':
                self.set_parameter('frequency_scale_factor',
                                   float(confs['frequency_scale_factor']))

            if conf_key == 'full_pp':
                if confs['full_pp'].lower() == '.true.':
                    self.set_parameter('is_full_pp', True)

            if conf_key == 'gamma_conversion_factor':
                self.set_parameter('gamma_conversion_factor',
                                   float(confs['gamma_conversion_factor']))

            if conf_key == 'grid_addresses':
                vals = [int(x) for x in
                        confs['grid_addresses'].replace(',', ' ').split()]
                if len(vals) % 3 == 0 and len(vals) > 0:
                    self.set_parameter('grid_addresses',
                                       np.reshape(vals, (-1, 3)))
                else:
                    self.setting_error("Grid addresses are incorrectly set.")


            if conf_key == 'grid_points':
                vals = [int(x) for x in
                        confs['grid_points'].replace(',', ' ').split()]
                self.set_parameter('grid_points', vals)

            if conf_key == 'ion_clamped':
                if confs['ion_clamped'].lower() == '.true.':
                    self.set_parameter('ion_clamped', True)

            if conf_key == 'bterta':
                if confs['bterta'].lower() == '.true.':
                    self.set_parameter('is_bterta', True)

            if conf_key == 'frequency_shift':
                if confs['frequency_shift'].lower() == '.true.':
                    self.set_parameter('is_frequency_shift', True)

            if conf_key == 'gruneisen':
                if confs['gruneisen'].lower() == '.true.':
                    self.set_parameter('is_gruneisen', True)

            if conf_key == 'imag_self_energy':
                if confs['imag_self_energy'].lower() == '.true.':
                    self.set_parameter('is_imag_self_energy', True)

            if conf_key == 'isotope':
                if confs['isotope'].lower() == '.true.':
                    self.set_parameter('is_isotope', True)

            if conf_key == 'joint_dos':
                if confs['joint_dos'].lower() == '.true.':
                    self.set_parameter('is_joint_dos', True)

            if conf_key == 'lbte':
                if confs['lbte'].lower() == '.true.':
                    self.set_parameter('is_lbte', True)

            if conf_key == 'linewidth':
                if confs['linewidth'].lower() == '.true.':
                    self.set_parameter('is_linewidth', True)

            if conf_key == 'N_U':
                if confs['N_U'].lower() == '.true.':
                    self.set_parameter('is_N_U', True)

            if conf_key == 'reducible_collision_matrix':
                if confs['reducible_collision_matrix'].lower() == '.true.':
                    self.set_parameter('is_reducible_collision_matrix', True)

            if conf_key == 'symmetrize_fc2':
                if confs['symmetrize_fc2'].lower() == '.true.':
                    self.set_parameter('is_symmetrize_fc2', True)

            if conf_key == 'symmetrize_fc3_q':
                if confs['symmetrize_fc3_q'].lower() == '.true.':
                    self.set_parameter('is_symmetrize_fc3_q', True)

            if conf_key == 'symmetrize_fc3_r':
                if confs['symmetrize_fc3_r'].lower() == '.true.':
                    self.set_parameter('is_symmetrize_fc3_r', True)

            if conf_key == 'mass_variances':
                vals = [fracval(x) for x in confs['mass_variances'].split()]
                if len(vals) < 1:
                    self.setting_error("Mass variance parameters are incorrectly set.")
                else:
                    self.set_parameter('mass_variances', vals)

            if conf_key == 'max_freepath':
                self.set_parameter('max_freepath', float(confs['max_freepath']))

            if conf_key == 'mesh_divisors':
                vals = [x for x in confs['mesh_divisors'].split()]
                if len(vals) == 3:
                    self.set_parameter('mesh_divisors', [int(x) for x in vals])
                elif len(vals) == 6:
                    divs = [int(x) for x in vals[:3]]
                    is_shift = [x.lower() == 't' for x in vals[3:]]
                    for i in range(3):
                        if is_shift[i] and (divs[i] % 2 != 0):
                            is_shift[i] = False
                            self.setting_error("Coarse grid shift along the " +
                                               ["first", "second", "third"][i] +
                                               " axis is not allowed.")
                    self.set_parameter('mesh_divisors', divs + is_shift)
                else:
                    self.setting_error("Mesh divisors are incorrectly set.")

            if conf_key == 'kappa_star':
                if confs['kappa_star'].lower() == '.false.':
                    self.set_parameter('is_kappa_star', False)

            if conf_key == 'pinv_cutoff':
                self.set_parameter('pinv_cutoff', float(confs['pinv_cutoff']))

            if conf_key == 'pinv_solver':
                self.set_parameter('pinv_solver', int(confs['pinv_solver']))

            if conf_key == 'pp_conversion_factor':
                self.set_parameter('pp_conversion_factor',
                                   float(confs['pp_conversion_factor']))

            if conf_key == 'read_amplitude':
                if confs['read_amplitude'].lower() == '.true.':
                    self.set_parameter('read_amplitude', True)

            if conf_key == 'read_collision':
                if confs['read_collision'] == 'all':
                    self.set_parameter('read_collision', 'all')
                else:
                    vals = [int(x) for x in confs['read_collision'].split()]
                    self.set_parameter('read_collision', vals)

            if conf_key == 'read_fc2':
                if confs['read_fc2'].lower() == '.true.':
                    self.set_parameter('read_fc2', True)

            if conf_key == 'read_fc3':
                if confs['read_fc3'].lower() == '.true.':
                    self.set_parameter('read_fc3', True)

            if conf_key == 'read_gamma':
                if confs['read_gamma'].lower() == '.true.':
                    self.set_parameter('read_gamma', True)

            if conf_key == 'read_phonon':
                if confs['read_phonon'].lower() == '.true.':
                    self.set_parameter('read_phonon', True)

            if conf_key == 'scattering_event_class':
                self.set_parameter('scattering_event_class',
                                   confs['scattering_event_class'])

            if conf_key == 'sigma_cutoff_width':
                self.set_parameter('sigma_cutoff_width',
                                   float(confs['sigma_cutoff_width']))

            if conf_key == 'temperatures':
                vals = [fracval(x) for x in confs['temperatures'].split()]
                if len(vals) < 1:
                    self.setting_error("Temperatures are incorrectly set.")
                else:
                    self.set_parameter('temperatures', vals)

            if conf_key == 'use_alm':
                if confs['use_alm'].lower() == '.true.':
                    self.set_parameter('use_alm', True)

            if conf_key == 'use_ave_pp':
                if confs['use_ave_pp'].lower() == '.true.':
                    self.set_parameter('use_ave_pp', True)

            if conf_key == 'write_amplitude':
                if confs['write_amplitude'].lower() == '.true.':
                    self.set_parameter('write_amplitude', True)

            if conf_key == 'write_gamma_detail':
                if confs['write_gamma_detail'].lower() == '.true.':
                    self.set_parameter('write_gamma_detail', True)

            if conf_key == 'write_gamma':
                if confs['write_gamma'].lower() == '.true.':
                    self.set_parameter('write_gamma', True)

            if conf_key == 'write_collision':
                if confs['write_collision'].lower() == '.true.':
                    self.set_parameter('write_collision', True)

            if conf_key == 'write_phonon':
                if confs['write_phonon'].lower() == '.true.':
                    self.set_parameter('write_phonon', True)

    def _set_settings(self):
        ConfParser.set_settings(self)
        params = self._parameters

        # Is getting least displacements?
        if 'create_displacements' in params:
            if params['create_displacements']:
                self._settings.set_create_displacements('displacements')

        # Supercell dimension for fc2
        if 'dim_fc2' in params:
            self._settings.set_phonon_supercell_matrix(params['dim_fc2'])

        # Boundary mean free path for thermal conductivity calculation
        if 'boundary_mfp' in params:
            self._settings.set_boundary_mfp(params['boundary_mfp'])

        # Peierls type approximation for squared ph-ph interaction strength
        if 'constant_averaged_pp_interaction' in params:
            self._settings.set_constant_averaged_pp_interaction(
                params['constant_averaged_pp_interaction'])

        # Cutoff distance of third-order force constants. Elements where any
        # pair of atoms has larger distance than cut-off distance are set zero.
        if 'cutoff_fc3_distance' in params:
            self._settings.set_cutoff_fc3_distance(params['cutoff_fc3_distance'])

        # Cutoff distance between pairs of displaced atoms used for supercell
        # creation with displacements and making third-order force constants
        if 'cutoff_pair_distance' in params:
            self._settings.set_cutoff_pair_distance(
                params['cutoff_pair_distance'])

        # This scale factor is multiplied to frequencies only, i.e., changes
        # frequencies but assumed not to change the physical unit
        if 'frequency_scale_factor' in params:
            self._settings.set_frequency_scale_factor(
                params['frequency_scale_factor'])

        # Gamma unit conversion factor
        if 'gamma_conversion_factor' in params:
            self._settings.set_gamma_conversion_factor(
                params['gamma_conversion_factor'])

        # Grid addresses (sets of three integer values)
        if 'grid_addresses' in params:
            self._settings.set_grid_addresses(params['grid_addresses'])

        # Grid points
        if 'grid_points' in params:
            self._settings.set_grid_points(params['grid_points'])

        # Atoms are clamped under applied strain in Gruneisen parameter calculation
        if 'ion_clamped' in params:
            self._settings.set_ion_clamped(params['ion_clamped'])

        # Calculate thermal conductivity in BTE-RTA
        if 'is_bterta' in params:
            self._settings.set_is_bterta(params['is_bterta'])

        # Calculate frequency_shifts
        if 'is_frequency_shift' in params:
            self._settings.set_is_frequency_shift(params['is_frequency_shift'])

        # Calculate full ph-ph interaction strength for RTA conductivity
        if 'is_full_pp' in params:
            self._settings.set_is_full_pp(params['is_full_pp'])

        # Calculate phonon-Gruneisen parameters
        if 'is_gruneisen' in params:
            self._settings.set_is_gruneisen(params['is_gruneisen'])

        # Calculate imaginary part of self energy
        if 'is_imag_self_energy' in params:
            self._settings.set_is_imag_self_energy(params['is_imag_self_energy'])

        # Calculate lifetime due to isotope scattering
        if 'is_isotope' in params:
            self._settings.set_is_isotope(params['is_isotope'])

        # Calculate joint-DOS
        if 'is_joint_dos' in params:
            self._settings.set_is_joint_dos(params['is_joint_dos'])

        # Calculate thermal conductivity in LBTE with Chaput's method
        if 'is_lbte' in params:
            self._settings.set_is_lbte(params['is_lbte'])

        # Calculate linewidths
        if 'is_linewidth' in params:
            self._settings.set_is_linewidth(params['is_linewidth'])

        # Calculate Normal and Umklapp processes
        if 'is_N_U' in params:
            self._settings.set_is_N_U(params['is_N_U'])

        # Solve reducible collision matrix but not reduced matrix
        if 'is_reducible_collision_matrix' in params:
            self._settings.set_is_reducible_collision_matrix(
                params['is_reducible_collision_matrix'])

        # Symmetrize fc2 by index exchange
        if 'is_symmetrize_fc2' in params:
            self._settings.set_is_symmetrize_fc2(params['is_symmetrize_fc2'])

        # Symmetrize phonon fc3 by index exchange
        if 'is_symmetrize_fc3_q' in params:
            self._settings.set_is_symmetrize_fc3_q(params['is_symmetrize_fc3_q'])

        # Symmetrize fc3 by index exchange
        if 'is_symmetrize_fc3_r' in params:
            self._settings.set_is_symmetrize_fc3_r(params['is_symmetrize_fc3_r'])

        # Mass variance parameters
        if 'mass_variances' in params:
            self._settings.set_mass_variances(params['mass_variances'])

        # Maximum mean free path
        if 'max_freepath' in params:
            self._settings.set_max_freepath(params['max_freepath'])

        # Divisors for mesh numbers
        if 'mesh_divisors' in params:
            self._settings.set_mesh_divisors(params['mesh_divisors'][:3])
            if len(params['mesh_divisors']) > 3:
                self._settings.set_coarse_mesh_shifts(
                    params['mesh_divisors'][3:])

        # Cutoff frequency for pseudo inversion of collision matrix
        if 'pinv_cutoff' in params:
            self._settings.set_pinv_cutoff(params['pinv_cutoff'])

        # Switch for pseudo-inverse solver
        if 'pinv_solver' in params:
            self._settings.set_pinv_solver(params['pinv_solver'])

        # Ph-ph interaction unit conversion factor
        if 'pp_conversion_factor' in params:
            self._settings.set_pp_conversion_factor(params['pp_conversion_factor'])

        # Read phonon-phonon interaction amplitudes from hdf5
        if 'read_amplitude' in params:
            self._settings.set_read_amplitude(params['read_amplitude'])

        # Read collision matrix and gammas from hdf5
        if 'read_collision' in params:
            self._settings.set_read_collision(params['read_collision'])

        # Read fc2 from hdf5
        if 'read_fc2' in params:
            self._settings.set_read_fc2(params['read_fc2'])

        # Read fc3 from hdf5
        if 'read_fc3' in params:
            self._settings.set_read_fc3(params['read_fc3'])

        # Read gammas from hdf5
        if 'read_gamma' in params:
            self._settings.set_read_gamma(params['read_gamma'])

        # Read phonons from hdf5
        if 'read_phonon' in params:
            self._settings.set_read_phonon(params['read_phonon'])

        # Sum partial kappa at q-stars
        if 'is_kappa_star' in params:
            self._settings.set_is_kappa_star(params['is_kappa_star'])

        # Scattering event class 1 or 2
        if 'scattering_event_class' in params:
            self._settings.set_scattering_event_class(
                params['scattering_event_class'])

        # Cutoff width of smearing function (ratio to sigma value)
        if 'sigma_cutoff_width' in params:
            self._settings.set_sigma_cutoff_width(params['sigma_cutoff_width'])

        # Temperatures
        if 'temperatures' in params:
            self._settings.set_temperatures(params['temperatures'])

        # Use ALM for creating force constants
        if 'use_alm' in params:
            self._settings.set_use_alm(params['use_alm'])

        # Use averaged ph-ph interaction
        if 'use_ave_pp' in params:
            self._settings.set_use_ave_pp(params['use_ave_pp'])

        # Write phonon-phonon interaction amplitudes to hdf5
        if 'write_amplitude' in params:
            self._settings.set_write_amplitude(params['write_amplitude'])

        # Write detailed imag-part of self energy to hdf5
        if 'write_gamma_detail' in params:
            self._settings.set_write_gamma_detail(
                params['write_gamma_detail'])

        # Write imag-part of self energy to hdf5
        if 'write_gamma' in params:
            self._settings.set_write_gamma(params['write_gamma'])

        # Write collision matrix and gammas to hdf5
        if 'write_collision' in params:
            self._settings.set_write_collision(params['write_collision'])

        # Write all phonons on grid points to hdf5
        if 'write_phonon' in params:
            self._settings.set_write_phonon(params['write_phonon'])



#!/usr/bin/env python

import numpy as np
from phonopy.interface.vasp import read_vasp
from phonopy.file_IO import parse_BORN
from phonopy.units import Bohr, Hartree
from phonopy.harmonic.force_constants import show_drift_force_constants
from phono3py.phonon3.fc3 import show_drift_fc3
from phono3py.phonon3 import Phono3py
from phono3py.file_IO import (parse_disp_fc3_yaml,
                              parse_disp_fc2_yaml,
                              parse_FORCES_FC2,
                              parse_FORCES_FC3,
                              read_fc3_from_hdf5,
                              read_fc2_from_hdf5)

cell = read_vasp("POSCAR-unitcell")
mesh = [11, 11, 11]
phono3py = Phono3py(cell,
                    np.diag([2, 2, 2]),
                    primitive_matrix=[[0, 0.5, 0.5],
                                      [0.5, 0, 0.5],
                                      [0.5, 0.5, 0]],
                    mesh=mesh,
                    log_level=1) # log_level=0 make phono3py quiet

# Create fc3 and fc2 from disp_fc3.yaml and FORCES_FC3
disp_dataset = parse_disp_fc3_yaml(filename="disp_fc3.yaml")
forces_fc3 = parse_FORCES_FC3(disp_dataset, filename="FORCES_FC3")
phono3py.produce_fc3(
    forces_fc3,
    displacement_dataset=disp_dataset,
    is_translational_symmetry=True,
    is_permutation_symmetry=True,
    is_permutation_symmetry_fc2=True)
fc3 = phono3py.get_fc3()
fc2 = phono3py.get_fc2()

# # Create fc2 from disp_fc2.yaml and FORCES_FC2
# disp_dataset2 = parse_disp_fc2_yaml(filename="disp_fc2.yaml")
# forces_fc2 = parse_FORCES_FC2(disp_dataset2, filename="FORCES_FC2")
# phono3py.produce_fc2(
#     forces_fc2,
#     displacement_dataset=disp_dataset2,
#     is_translational_symmetry=True,
#     is_permutation_symmetry=True)

# # Read fc3 and fc2 from c3.hdf5 and fc2.hdf5
# fc3 = read_fc3_from_hdf5(filename="fc3.hdf5")
# fc2 = read_fc2_from_hdf5(filename="fc2.hdf5")
# phono3py.set_fc3(fc3)
# phono3py.set_fc2(fc2)

show_drift_fc3(fc3)
show_drift_force_constants(fc2, name='fc2')

# # For special cases like NAC
# primitive = phono3py.get_phonon_primitive()
# nac_params = parse_BORN(primitive, filename="BORN")
# nac_params['factor'] = Hartree * Bohr
# phono3py.set_phph_interaction(nac_params=nac_params)

phono3py.run_thermal_conductivity(temperatures=range(0, 1001, 10),
                                  write_kappa=True)

# Conductivity_RTA object (https://git.io/vVRUW)
cond_rta = phono3py.get_thermal_conductivity()




import numpy
import platform
import os

try:
    from setuptools import setup, Extension
    use_setuptools = True
    print("setuptools is used.")
except ImportError:
    from distutils.core import setup, Extension
    use_setuptools = False
    print("distutils is used.")

include_dirs_numpy = [numpy.get_include()]
extra_link_args = []
cc = None
if 'CC' in os.environ:
    if 'clang' in os.environ['CC']:
        cc = 'clang'
    if 'gcc' in os.environ['CC']:
        cc = 'gcc'
if cc == 'gcc' or cc is None:
    extra_link_args.append('-lgomp')

# Workaround Python issue 21121
import sysconfig
config_var = sysconfig.get_config_var("CFLAGS")
if config_var is not None and "-Werror=declaration-after-statement" in config_var:
    os.environ['CFLAGS'] = config_var.replace(
        "-Werror=declaration-after-statement", "")    

sources = ['c/_phono3py.c',
           'c/harmonic/dynmat.c',
           'c/harmonic/lapack_wrapper.c',
           'c/harmonic/phonoc_array.c',
           'c/harmonic/phonoc_utils.c',
           'c/anharmonic/phonon3/fc3.c',
           'c/anharmonic/phonon3/frequency_shift.c',
           'c/anharmonic/phonon3/interaction.c',
           'c/anharmonic/phonon3/real_to_reciprocal.c',
           'c/anharmonic/phonon3/reciprocal_to_normal.c',
           'c/anharmonic/phonon3/imag_self_energy_with_g.c',
           'c/anharmonic/phonon3/pp_collision.c',
           'c/anharmonic/phonon3/collision_matrix.c',
           'c/anharmonic/other/isotope.c',
           'c/anharmonic/triplet/triplet.c',
           'c/anharmonic/triplet/triplet_kpoint.c',
           'c/anharmonic/triplet/triplet_iw.c',
           'c/spglib/mathfunc.c',
           'c/spglib/kpoint.c',
           'c/kspclib/kgrid.c',
           'c/kspclib/tetrahedron_method.c']

extra_compile_args = ['-fopenmp',]
include_dirs = ['c/harmonic_h',
                'c/anharmonic_h',
                'c/spglib_h',
                'c/kspclib_h'] + include_dirs_numpy
library_dirs = []
define_macros = []

extra_link_args_lapacke = []
include_dirs_lapacke = []
library_dirs_lapacke = []

# C macro definitions:
# - MULTITHREADED_BLAS
#   This deactivates OpenMP multithread harmonic phonon calculation,
#   since inside each phonon calculation, zheev is called.
#   When using multithread BLAS, this macro has to be set and
#   by this all phonons on q-points should be calculated in series.
# - MKL_LAPACKE:
#   This sets definitions and functions needed when using MKL lapacke.
#   Phono3py complex values are handled based on those provided by Netlib
#   lapacke. However MKL lapacke doesn't provide some macros and functions
#   that provided Netlib. This macro defines those used in phono3py among them.
if os.path.isfile("mkl.py"):
    # This supposes that MKL multithread BLAS is used.
    # This is invoked when mkl.py exists on the current directory.

    #### Example of mkl.py ####
    # extra_link_args_lapacke += ['-L/opt/intel/mkl/lib/intel64',
    #                            '-lmkl_intel_ilp64', '-lmkl_intel_thread',
    #                            '-lmkl_core']
    # library_dirs_lapacke += []
    # include_dirs_lapacke += ['/opt/intel/mkl/include']

    print("MKL LAPACKE is to be used.")
    from mkl import (extra_link_args_lapacke, include_dirs_lapacke,
                     library_dirs_lapacke)
    if use_setuptools:
        extra_compile_args += ['-DMKL_LAPACKE',
                               '-DMULTITHREADED_BLAS']
    else:
        define_macros += [('MKL_LAPACKE', None),
                          ('MULTITHREADED_BLAS', None)]
elif os.path.isfile("libopenblas.py"):
    # This supposes that multithread openBLAS is used.
    # This is invoked when libopenblas.py exists on the current directory.

    #### Example of libopenblas.py ####
    # extra_link_args_lapacke += ['-lopenblas']

    from libopenblas import extra_link_args_lapacke
    include_dirs_lapacke += []
    library_dirs_lapacke += []
    if use_setuptools:
        extra_compile_args += ['-DMULTITHREADED_BLAS']
    else:
        define_macros += [('MULTITHREADED_BLAS', None)]
elif (platform.system() == 'Darwin' and
      os.path.isfile('/opt/local/lib/libopenblas.a')):
    # This supposes lapacke with single-thread openBLAS provided by MacPort is
    # used.
    # % sudo port install gcc6
    # % sudo port select --set gcc mp-gcc
    # % sudo port install OpenBLAS +gcc6
    extra_link_args_lapacke += ['/opt/local/lib/libopenblas.a']
    include_dirs_lapacke += ['/opt/local/include']
    library_dirs_lapacke += []
elif os.path.isfile('/usr/lib/liblapacke.so'):
    # This supposes that lapacke with single-thread BLAS is installed on system.
    extra_link_args_lapacke += ['-llapacke', '-llapack', '-lblas']
    include_dirs_lapacke += []
    library_dirs_lapacke += []
else:
    # Here is the default lapacke linkage setting.
    # Please modify according to your system environment.
    # Without using multithreaded BLAS, DMULTITHREADED_BLAS is better to be
    # removed to activate OpenMP multithreading harmonic phonon calculation, but
    # this is not mandatory.
    #
    # The below supposes that lapacke with multithread openblas is used.
    # Even if using single-thread BLAS and deactivating OpenMP
    # multithreading for harmonic phonon calculation, the total performance
    # decrease is considered marginal.
    #
    # For conda: Try installing with dynamic link library of openblas by
    # % conda install numpy scipy h5py pyyaml matplotlib openblas
    extra_link_args_lapacke += ['-lopenblas']
    include_dirs_lapacke += []
    library_dirs_lapacke += []
    if use_setuptools:
        extra_compile_args += ['-DMULTITHREADED_BLAS']
    else:
        define_macros += [('MULTITHREADED_BLAS', None)]

## Uncomment below to measure reciprocal_to_normal_squared_openmp performance
# define_macros += [('MEASURE_R2N', None)]

extra_link_args += extra_link_args_lapacke
include_dirs += include_dirs_lapacke
library_dirs += library_dirs_lapacke
extension_phono3py = Extension(
    'phono3py._phono3py',
    include_dirs=include_dirs,
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    library_dirs=library_dirs,
    define_macros=define_macros,
    sources=sources)

packages_phono3py = ['phono3py',
                     'phono3py.cui',
                     'phono3py.other',
                     'phono3py.phonon',
                     'phono3py.phonon3']
scripts_phono3py = ['scripts/phono3py',
                    'scripts/phono3py-kaccum',
                    'scripts/phono3py-kdeplot']

## This is for the test of libflame
##
# use_libflame = False
# if use_libflame:
#     sources.append('c/anharmonic/flame_wrapper.c')
#     extra_link_args.append('../libflame-bin/lib/libflame.a')
#     include_dirs_libflame = ['../libflame-bin/include']
#     include_dirs += include_dirs_libflame

########################
# _lapackepy extension #
########################
include_dirs_lapackepy = ['c/harmonic_h',] + include_dirs_numpy
sources_lapackepy = ['c/_lapackepy.c',
                     'c/harmonic/dynmat.c',
                     'c/harmonic/phonon.c',
                     'c/harmonic/phonoc_array.c',
                     'c/harmonic/phonoc_utils.c',
                     'c/harmonic/lapack_wrapper.c']
extension_lapackepy = Extension(
    'phono3py._lapackepy',
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    include_dirs=include_dirs,
    sources=sources_lapackepy)

if __name__ == '__main__':
    version_nums = [None, None, None]
    with open("phono3py/version.py") as w:
        for line in w:
            if "__version__" in line:
                for i, num in enumerate(line.split()[2].strip('\"').split('.')):
                    version_nums[i] = int(num)

    # To deploy to pypi by travis-CI
    if os.path.isfile("__nanoversion__.txt"):
        with open('__nanoversion__.txt') as nv:
            nanoversion = 0
            try :
                for line in nv:
                    nanoversion = int(line.strip())
                    break
            except ValueError :
                nanoversion = 0
            if nanoversion:
                version_nums.append(nanoversion)

    if None in version_nums:
        print("Failed to get version number in setup.py.")
        raise

    version_number = ".".join(["%d" % n for n in version_nums])
    if use_setuptools:
        setup(name='phono3py',
              version=version_number,
              description='This is the phono3py module.',
              author='Atsushi Togo',
              author_email='atz.togo@gmail.com',
              url='http://atztogo.github.io/phono3py/',
              packages=packages_phono3py,
              requires=['numpy', 'PyYAML', 'matplotlib', 'h5py', 'phonopy'],
              provides=['phono3py'],
              scripts=scripts_phono3py,
              ext_modules=[extension_lapackepy, extension_phono3py])
    else:
        setup(name='phono3py',
              version=version_number,
              description='This is the phono3py module.',
              author='Atsushi Togo',
              author_email='atz.togo@gmail.com',
              url='http://atztogo.github.io/phono3py/',
              packages=packages_phono3py,
              requires=['numpy', 'PyYAML', 'matplotlib', 'h5py', 'phonopy'],
              provides=['phono3py'],
              scripts=scripts_phono3py,
              ext_modules=[extension_lapackepy, extension_phono3py])



import sys
import numpy as np
from phonopy.harmonic.force_constants import (similarity_transformation,
                                              set_permutation_symmetry,
                                              distribute_force_constants,
                                              solve_force_constants,
                                              get_rotated_displacement,
                                              get_positions_sent_by_rot_inv,
                                              set_translational_invariance)
from phono3py.phonon3.displacement_fc3 import (get_reduced_site_symmetry,
                                               get_bond_symmetry)
from phonopy.structure.cells import get_equivalent_smallest_vectors

def get_fc3(supercell,
            disp_dataset,
            fc2,
            symmetry,
            translational_symmetry_type=0,
            is_permutation_symmetry=False,
            verbose=False):
    num_atom = supercell.get_number_of_atoms()
    fc3_least_atoms = np.zeros((num_atom, num_atom, num_atom, 3, 3, 3),
                               dtype='double')

    if 'cutoff_distance' in disp_dataset:
        _get_fc3_least_atoms(fc3_least_atoms,
                             supercell,
                             disp_dataset,
                             fc2,
                             symmetry,
                             False,
                             False,
                             verbose)
    else:
        _get_fc3_least_atoms(fc3_least_atoms,
                             supercell,
                             disp_dataset,
                             fc2,
                             symmetry,
                             translational_symmetry_type,
                             is_permutation_symmetry,
                             verbose)

    if verbose:
        print("Expanding fc3")

    first_disp_atoms = np.unique(
        [x['number'] for x in disp_dataset['first_atoms']])
    rotations = symmetry.get_symmetry_operations()['rotations']
    translations = symmetry.get_symmetry_operations()['translations']
    symprec = symmetry.get_symmetry_tolerance()
    lattice = supercell.get_cell().T
    positions = supercell.get_scaled_positions()

    fc3 = distribute_fc3(fc3_least_atoms,
                         first_disp_atoms,
                         lattice,
                         positions,
                         rotations,
                         translations,
                         symprec,
                         overwrite=False,
                         verbose=verbose)

    if 'cutoff_distance' in disp_dataset:
        if verbose:
            print("Cutting-off fc3 (cut-off distance: %f)" %
                  disp_dataset['cutoff_distance'])
        cutoff_fc3(fc3,
                   supercell,
                   disp_dataset,
                   symmetry,
                   verbose=verbose)
        if translational_symmetry_type:
            if verbose:
                print("Imposing translational invariance symmetry to fc3")
            set_translational_invariance_fc3(
                fc3, translational_symmetry_type=translational_symmetry_type)
        if is_permutation_symmetry:
            if verbose:
                print("Imposing index permulation symmetry to fc3")
            set_permutation_symmetry_fc3(fc3)
    else:
        if translational_symmetry_type:
            if verbose:
                print("Imposing translational invariance symmetry to fc3")
            set_translational_invariance_fc3_per_index(
                fc3, translational_symmetry_type=translational_symmetry_type)
        if is_permutation_symmetry:
            if verbose:
                print("Imposing index permulation symmetry to fc3")
            set_permutation_symmetry_fc3(fc3)

    return fc3


def distribute_fc3(fc3_least_atoms,
                   first_disp_atoms,
                   lattice,
                   positions,
                   rotations,
                   translations,
                   symprec,
                   overwrite=True,
                   verbose=False):
    num_atom = len(positions)
    atom_mapping = np.zeros(num_atom, dtype='intc')

    if overwrite:
        fc3 = fc3_least_atoms
    else:
        fc3 = np.zeros((num_atom, num_atom, num_atom, 3, 3, 3), dtype='double')

    for i in range(num_atom):
        # if i in first_disp_atoms:
        #     continue

        for atom_index_done in first_disp_atoms:
            rot_num = _get_atom_mapping_by_symmetry(lattice,
                                                    positions,
                                                    i,
                                                    atom_index_done,
                                                    rotations,
                                                    translations,
                                                    symprec)
            if rot_num > -1:
                i_rot = atom_index_done
                rot = rotations[rot_num]
                trans = translations[rot_num]
                break

        if rot_num < 0:
            print("Position or symmetry may be wrong.")
            raise ValueError

        for j in range(num_atom):
            atom_mapping[j] = _get_atom_by_symmetry(lattice,
                                                    positions,
                                                    rot,
                                                    trans,
                                                    j,
                                                    symprec)

        rot_cart_inv = np.array(similarity_transformation(lattice, rot).T,
                                dtype='double', order='C')

        if not (overwrite and i == i_rot):
            if verbose > 2:
                print("    [ %d, x, x ] to [ %d, x, x ]" % (i_rot + 1, i + 1))
                sys.stdout.flush()

            try:
                import phono3py._phono3py as phono3c
                phono3c.distribute_fc3(fc3,
                                       fc3_least_atoms,
                                       i,
                                       atom_mapping,
                                       rot_cart_inv)

            except ImportError:
                for j in range(num_atom):
                    j_rot = atom_mapping[j]
                    for k in range(num_atom):
                        k_rot = atom_mapping[k]
                        fc3[i, j, k] = third_rank_tensor_rotation(
                            rot_cart_inv, fc3_least_atoms[i_rot, j_rot, k_rot])

    if not overwrite:
        return fc3

def set_permutation_symmetry_fc3(fc3):
    try:
        import phono3py._phono3py as phono3c
        phono3c.permutation_symmetry_fc3(fc3)
    except ImportError:
        num_atom = fc3.shape[0]
        for i in range(num_atom):
            for j in range(i, num_atom):
                for k in range(j, num_atom):
                    fc3_elem = set_permutation_symmetry_fc3_elem(fc3, i, j, k)
                    copy_permutation_symmetry_fc3_elem(fc3, fc3_elem, i, j, k)

def set_permutation_symmetry_fc3_deprecated(fc3):
    fc3_sym = np.zeros(fc3.shape, dtype='double')
    for (i, j, k) in list(np.ndindex(fc3.shape[:3])):
        fc3_sym[i, j, k] = set_permutation_symmetry_fc3_elem(fc3, i, j, k)

    for (i, j, k) in list(np.ndindex(fc3.shape[:3])):
        fc3[i, j, k] = fc3_sym[i, j, k]

def copy_permutation_symmetry_fc3_elem(fc3, fc3_elem, a, b, c):
    for (i, j, k) in list(np.ndindex(3, 3, 3)):
        fc3[a, b, c, i, j, k] = fc3_elem[i, j, k]
        fc3[c, a, b, k, i, j] = fc3_elem[i, j, k]
        fc3[b, c, a, j, k, i] = fc3_elem[i, j, k]
        fc3[a, c, b, i, k, j] = fc3_elem[i, j, k]
        fc3[b, a, c, j, i, k] = fc3_elem[i, j, k]
        fc3[c, b, a, k, j, i] = fc3_elem[i, j, k]

def set_permutation_symmetry_fc3_elem(fc3, a, b, c, divisor=6):
    tensor3 = np.zeros((3, 3, 3), dtype='double')
    for (i, j, k) in list(np.ndindex(3, 3, 3)):
        tensor3[i, j, k] = (fc3[a, b, c, i, j, k] +
                            fc3[c, a, b, k, i, j] +
                            fc3[b, c, a, j, k, i] +
                            fc3[a, c, b, i, k, j] +
                            fc3[b, a, c, j, i, k] +
                            fc3[c, b, a, k, j, i]) / divisor
    return tensor3

def set_translational_invariance_fc3(fc3,
                                     translational_symmetry_type=1):
    for i in range(3):
        set_translational_invariance_fc3_per_index(
            fc3,
            index=i,
            translational_symmetry_type=translational_symmetry_type)

def set_translational_invariance_fc3_per_index(fc3,
                                               index=0,
                                               translational_symmetry_type=1):
    for i in range(fc3.shape[(1 + index) % 3]):
        for j in range(fc3.shape[(2 + index) % 3]):
            for k, l, m in list(np.ndindex(3, 3, 3)):
                if translational_symmetry_type == 2: # Type 2
                    if index == 0:
                        fc_abs = np.abs(fc3[:, i, j, k, l, m])
                        fc_sum = np.sum(fc3[:, i, j, k, l, m])
                        fc_abs_sum = np.sum(fc_abs)
                        fc3[:, i, j, k, l, m] -= (
                            fc_sum / fc_abs_sum * fc_abs)
                    elif index == 1:
                        fc_abs = np.abs(fc3[i, :, j, k, l, m])
                        fc_sum = np.sum(fc3[i, :, j, k, l, m])
                        fc_abs_sum = np.sum(fc_abs)
                        fc3[i, :, j, k, l, m] -= (
                            fc_sum / fc_abs_sum * fc_abs)
                    elif index == 2:
                        fc_abs = np.abs(fc3[i, j, :, k, l, m])
                        fc_sum = np.sum(fc3[i, j, :, k, l, m])
                        fc_abs_sum = np.sum(fc_abs)
                        fc3[i, j, :, k, l, m] -= (
                            fc_sum / fc_abs_sum * fc_abs)
                else: # Type 1
                    if index == 0:
                        fc3[:, i, j, k, l, m] -= np.sum(
                            fc3[:, i, j, k, l, m]) / fc3.shape[0]
                    elif index == 1:
                        fc3[j, :, i, k, l, m] -= np.sum(
                            fc3[j, :, i, k, l, m]) / fc3.shape[1]
                    elif index == 2:
                        fc3[i, j, :, k, l, m] -= np.sum(
                            fc3[i, j, :, k, l, m]) / fc3.shape[2]

def third_rank_tensor_rotation(rot_cart, tensor):
    rot_tensor = np.zeros((3,3,3), dtype='double')
    for i in (0,1,2):
        for j in (0,1,2):
            for k in (0,1,2):
                rot_tensor[i, j, k] = _third_rank_tensor_rotation_elem(
                    rot_cart, tensor, i, j, k)
    return rot_tensor


def get_delta_fc2(dataset_second_atoms,
                  atom1,
                  fc2,
                  supercell,
                  reduced_site_sym,
                  translational_symmetry_type,
                  is_permutation_symmetry,
                  symprec):
    disp_fc2 = get_constrained_fc2(supercell,
                                   dataset_second_atoms,
                                   atom1,
                                   reduced_site_sym,
                                   translational_symmetry_type,
                                   is_permutation_symmetry,
                                   symprec)
    return disp_fc2 - fc2

def get_constrained_fc2(supercell,
                        dataset_second_atoms,
                        atom1,
                        reduced_site_sym,
                        translational_symmetry_type,
                        is_permutation_symmetry,
                        symprec):
    """
    dataset_second_atoms: [{'number': 7,
                            'displacement': [],
                            'delta_forces': []}, ...]
    """
    lattice = supercell.get_cell().T
    positions = supercell.get_scaled_positions()
    num_atom = supercell.get_number_of_atoms()

    fc2 = np.zeros((num_atom, num_atom, 3, 3), dtype='double')
    atom_list = np.unique([x['number'] for x in dataset_second_atoms])
    for atom2 in atom_list:
        disps2 = []
        sets_of_forces = []
        for disps_second in dataset_second_atoms:
            if atom2 != disps_second['number']:
                continue
            bond_sym = get_bond_symmetry(
                reduced_site_sym,
                lattice,
                positions,
                atom1,
                atom2,
                symprec)

            disps2.append(disps_second['displacement'])
            sets_of_forces.append(disps_second['delta_forces'])

        solve_force_constants(fc2,
                              atom2,
                              disps2,
                              sets_of_forces,
                              supercell,
                              bond_sym,
                              symprec)

    # Shift positions according to set atom1 is at origin
    pos_center = positions[atom1].copy()
    positions -= pos_center
    distribute_force_constants(fc2,
                               range(num_atom),
                               atom_list,
                               lattice,
                               positions,
                               np.array(reduced_site_sym,
                                        dtype='intc', order='C'),
                               np.zeros((len(reduced_site_sym), 3),
                                        dtype='double'),
                               symprec)

    if translational_symmetry_type:
        set_translational_invariance(
            fc2,
            translational_symmetry_type=translational_symmetry_type)

    if is_permutation_symmetry:
        set_permutation_symmetry(fc2)

    return fc2


def solve_fc3(fc3,
              first_atom_num,
              supercell,
              site_symmetry,
              displacements_first,
              delta_fc2s,
              symprec,
              pinv_solver="numpy",
              verbose=False):

    if pinv_solver == "numpy":
        solver = "numpy.linalg.pinv"
    else:
        try:
            import phono3py._lapackepy as lapackepy
            solver = "lapacke-dgesvd"
        except ImportError:
            solver = "numpy"

    if verbose:
        text = ("Computing fc3[ %d, x, x ] using %s with " %
                (first_atom_num + 1, solver))
        if len(displacements_first) > 1:
            text += "displacements:"
        else:
            text += "a displacement:"
        print(text)
        for i, v in enumerate(displacements_first):
            print("    [%7.4f %7.4f %7.4f]" % tuple(v))
            sys.stdout.flush()
        if verbose > 2:
            print("  Site symmetry:")
            for i, v in enumerate(site_symmetry):
                print("    [%2d %2d %2d] #%2d" % tuple(list(v[0])+[i + 1]))
                print("    [%2d %2d %2d]" % tuple(v[1]))
                print("    [%2d %2d %2d]\n" % tuple(v[2]))
                sys.stdout.flush()

    lattice = supercell.get_cell().T
    site_sym_cart = [similarity_transformation(lattice, sym)
                     for sym in site_symmetry]
    num_atom = supercell.get_number_of_atoms()
    positions = supercell.get_scaled_positions()
    pos_center = positions[first_atom_num].copy()
    positions -= pos_center
    rot_map_syms = get_positions_sent_by_rot_inv(lattice,
                                                 positions,
                                                 site_symmetry,
                                                 symprec)

    rot_disps = get_rotated_displacement(displacements_first, site_sym_cart)

    if "numpy" in solver:
        inv_U = np.linalg.pinv(rot_disps)
    else:
        inv_U = np.zeros((rot_disps.shape[1], rot_disps.shape[0]),
                         dtype='double')
        lapackepy.pinv(inv_U, rot_disps, 1e-13)

    for (i, j) in list(np.ndindex(num_atom, num_atom)):
        fc3[first_atom_num, i, j] = np.dot(inv_U, _get_rotated_fc2s(
                i, j, delta_fc2s, rot_map_syms, site_sym_cart)).reshape(3, 3, 3)

def cutoff_fc3(fc3,
               supercell,
               disp_dataset,
               symmetry,
               verbose=False):
    if verbose:
        print("Building atom mapping table...")
    fc3_done = _get_fc3_done(supercell, disp_dataset, symmetry)

    if verbose:
        print("Creating contracted fc3...")
    num_atom = supercell.get_number_of_atoms()
    for i in range(num_atom):
        for j in range(i, num_atom):
            for k in range(j, num_atom):
                ave_fc3 = _set_permutation_symmetry_fc3_elem_with_cutoff(
                    fc3, fc3_done, i, j, k)
                copy_permutation_symmetry_fc3_elem(fc3, ave_fc3, i, j, k)

def cutoff_fc3_by_zero(fc3, supercell, cutoff_distance, symprec=1e-5):
    num_atom = supercell.get_number_of_atoms()
    lattice = supercell.get_cell().T
    min_distances = np.zeros((num_atom, num_atom), dtype='double')
    for i in range(num_atom): # run in supercell
        for j in range(num_atom): # run in primitive
            min_distances[i, j] = np.linalg.norm(
                np.dot(lattice,
                       get_equivalent_smallest_vectors(
                           i, j, supercell, lattice.T, symprec)[0]))

    for i, j, k in np.ndindex(num_atom, num_atom, num_atom):
        for pair in ((i, j), (j, k), (k, i)):
            if min_distances[pair] > cutoff_distance:
                fc3[i, j, k] = 0
                break

def show_drift_fc3(fc3, name="fc3"):
    num_atom = fc3.shape[0]
    maxval1 = 0
    maxval2 = 0
    maxval3 = 0
    klm1 = [0, 0, 0]
    klm2 = [0, 0, 0]
    klm3 = [0, 0, 0]
    for i, j, k, l, m in list(np.ndindex((num_atom, num_atom, 3, 3, 3))):
        val1 = fc3[:, i, j, k, l, m].sum()
        val2 = fc3[i, :, j, k, l, m].sum()
        val3 = fc3[i, j, :, k, l, m].sum()
        if abs(val1) > abs(maxval1):
            maxval1 = val1
            klm1 = [k, l, m]
        if abs(val2) > abs(maxval2):
            maxval2 = val2
            klm2 = [k, l, m]
        if abs(val3) > abs(maxval3):
            maxval3 = val3
            klm3 = [k, l, m]
    text = "max drift of %s: " % name
    text += "%f (%s%s%s) " % (maxval1,
                              "xyz"[klm1[0]], "xyz"[klm1[1]], "xyz"[klm1[2]])
    text += "%f (%s%s%s) " % (maxval2,
                              "xyz"[klm2[0]], "xyz"[klm2[1]], "xyz"[klm2[2]])
    text += "%f (%s%s%s)" % (maxval3,
                             "xyz"[klm3[0]], "xyz"[klm3[1]], "xyz"[klm3[2]])
    print(text)

def _set_permutation_symmetry_fc3_elem_with_cutoff(fc3, fc3_done, a, b, c):
    sum_done = (fc3_done[a, b, c] +
                fc3_done[c, a, b] +
                fc3_done[b, c, a] +
                fc3_done[b, a, c] +
                fc3_done[c, b, a] +
                fc3_done[a, c, b])
    tensor3 = np.zeros((3, 3, 3), dtype='double')
    if sum_done > 0:
        for (i, j, k) in list(np.ndindex(3, 3, 3)):
            tensor3[i, j, k] = (fc3[a, b, c, i, j, k] * fc3_done[a, b, c] +
                                fc3[c, a, b, k, i, j] * fc3_done[c, a, b] +
                                fc3[b, c, a, j, k, i] * fc3_done[b, c, a] +
                                fc3[a, c, b, i, k, j] * fc3_done[a, c, b] +
                                fc3[b, a, c, j, i, k] * fc3_done[b, a, c] +
                                fc3[c, b, a, k, j, i] * fc3_done[c, b, a])
            tensor3[i, j, k] /= sum_done
    return tensor3

def _get_fc3_least_atoms(fc3,
                         supercell,
                         disp_dataset,
                         fc2,
                         symmetry,
                         translational_symmetry_type,
                         is_permutation_symmetry,
                         verbose):
    symprec = symmetry.get_symmetry_tolerance()
    unique_first_atom_nums = np.unique(
        [x['number'] for x in disp_dataset['first_atoms']])
    for first_atom_num in unique_first_atom_nums:
        _get_fc3_one_atom(fc3,
                          supercell,
                          disp_dataset,
                          fc2,
                          first_atom_num,
                          symmetry.get_site_symmetry(first_atom_num),
                          translational_symmetry_type,
                          is_permutation_symmetry,
                          symprec,
                          verbose)

def _get_fc3_one_atom(fc3,
                      supercell,
                      disp_dataset,
                      fc2,
                      first_atom_num,
                      site_symmetry,
                      translational_symmetry_type,
                      is_permutation_symmetry,
                      symprec,
                      verbose):
    displacements_first = []
    delta_fc2s = []
    for dataset_first_atom in disp_dataset['first_atoms']:
        if first_atom_num != dataset_first_atom['number']:
            continue

        displacements_first.append(dataset_first_atom['displacement'])
        if 'delta_fc2' in dataset_first_atom:
            delta_fc2s.append(dataset_first_atom['delta_fc2'])
        else:
            direction = np.dot(dataset_first_atom['displacement'],
                               np.linalg.inv(supercell.get_cell()))
            reduced_site_sym = get_reduced_site_symmetry(
                site_symmetry, direction, symprec)
            delta_fc2s.append(get_delta_fc2(
                    dataset_first_atom['second_atoms'],
                    dataset_first_atom['number'],
                    fc2,
                    supercell,
                    reduced_site_sym,
                    translational_symmetry_type,
                    is_permutation_symmetry,
                    symprec))

    solve_fc3(fc3,
              first_atom_num,
              supercell,
              site_symmetry,
              displacements_first,
              delta_fc2s,
              symprec,
              verbose=verbose)



def _get_rotated_fc2s(i, j, fc2s, rot_map_syms, site_sym_cart):
    num_sym = len(site_sym_cart)
    rotated_fc2s = []
    for fc2 in fc2s:
        for sym, map_sym in zip(site_sym_cart, rot_map_syms):
            fc2_rot = fc2[map_sym[i], map_sym[j]]
            rotated_fc2s.append(similarity_transformation(sym, fc2_rot))
    return np.reshape(rotated_fc2s, (-1, 9))

def _third_rank_tensor_rotation_elem(rot, tensor, l, m, n):
    sum_elems = 0.
    for i in (0, 1, 2):
        for j in (0, 1, 2):
            for k in (0, 1, 2):
                sum_elems += rot[l, i] * rot[m, j] * rot[n, k] * tensor[i, j, k]
    return sum_elems

def _get_fc3_done(supercell, disp_dataset, symmetry):
    num_atom = supercell.get_number_of_atoms()
    fc3_done = np.zeros((num_atom, num_atom, num_atom), dtype='byte')
    symprec = symmetry.get_symmetry_tolerance()
    lattice = supercell.get_cell().T
    positions = supercell.get_scaled_positions()
    rotations = symmetry.get_symmetry_operations()['rotations']
    translations = symmetry.get_symmetry_operations()['translations']

    atom_mapping = []
    for rot, trans in zip(rotations, translations):
        atom_indices = []
        for rot_pos in (np.dot(positions, rot.T) + trans):
            diffs = positions - rot_pos
            diffs -= np.rint(diffs)
            diffs = np.dot(diffs, lattice.T)
            dists = np.sqrt(np.sum(diffs ** 2, axis=1))
            atom_indices.append((np.where(dists < symprec))[0][0])
        atom_mapping.append(atom_indices)

    for dataset_first_atom in disp_dataset['first_atoms']:
        first_atom_num = dataset_first_atom['number']
        site_symmetry = symmetry.get_site_symmetry(first_atom_num)
        direction = np.dot(dataset_first_atom['displacement'],
                           np.linalg.inv(supercell.get_cell()))
        reduced_site_sym = get_reduced_site_symmetry(
            site_symmetry, direction, symprec)
        least_second_atom_nums = []
        for second_atoms in dataset_first_atom['second_atoms']:
            if second_atoms['included']:
                least_second_atom_nums.append(second_atoms['number'])
        positions_shifted = positions - positions[first_atom_num]
        least_second_atom_nums = np.unique(least_second_atom_nums)

        second_atom_nums = []
        for red_rot in reduced_site_sym:
            for i in least_second_atom_nums:
                second_atom_nums.append(
                    _get_atom_by_symmetry(lattice,
                                          positions_shifted,
                                          red_rot,
                                          np.zeros(3, dtype='double'),
                                          i,
                                          symprec))
        second_atom_nums = np.unique(second_atom_nums)

        for i in range(len(rotations)):
            rotated_atom1 = atom_mapping[i][first_atom_num]
            for j in second_atom_nums:
                fc3_done[rotated_atom1, atom_mapping[i][j]] = 1

    return fc3_done

def _get_atom_by_symmetry(lattice,
                          positions,
                          rotation,
                          trans,
                          atom_number,
                          symprec=1e-5):
    rot_pos = np.dot(positions[atom_number], rotation.T) + trans
    for i, pos in enumerate(positions):
        diff = pos - rot_pos
        diff -= np.rint(diff)
        dist = np.linalg.norm(np.dot(lattice, diff))
        if dist < symprec:
            return i

    print("Position or symmetry is wrong.")
    raise ValueError

def _get_atom_mapping_by_symmetry(lattice,
                                  positions,
                                  atom_search,
                                  atom_target,
                                  rotations,
                                  translations,
                                  symprec):
    map_sym = -1

    for i, (r, t) in enumerate(zip(rotations, translations)):
        rot_pos = np.dot(positions[atom_search], r.T) + t
        diff = rot_pos - positions[atom_target]
        diff -= np.rint(diff)
        dist = np.linalg.norm(np.dot(lattice, diff))
        if dist < symprec:
            map_sym = i
            break

    return map_sym




import numpy as np

class ReciprocalToNormal(object):
    def __init__(self,
                 primitive,
                 frequencies,
                 eigenvectors,
                 band_indices,
                 cutoff_frequency=0):
        self._primitive = primitive
        self._frequencies = frequencies
        self._eigenvectors = eigenvectors
        self._band_indices = band_indices
        self._cutoff_frequency = cutoff_frequency

        self._masses = self._primitive.get_masses()

        self._fc3_normal = None
        self._fc3_reciprocal = None

    def run(self, fc3_reciprocal, grid_triplet):
        num_band = self._primitive.get_number_of_atoms() * 3
        self._fc3_reciprocal = fc3_reciprocal
        dtype = "c%d" % (np.dtype('double').itemsize * 2)
        self._fc3_normal = np.zeros(
            (len(self._band_indices), num_band, num_band), dtype=dtype)
        self._reciprocal_to_normal(grid_triplet)

    def get_reciprocal_to_normal(self):
        return self._fc3_normal

    def _reciprocal_to_normal(self, grid_triplet):
        e1, e2, e3 = self._eigenvectors[grid_triplet]
        f1, f2, f3 = self._frequencies[grid_triplet]
        num_band = len(f1)
        cutoff = self._cutoff_frequency
        for (i, j, k) in list(np.ndindex(
                len(self._band_indices), num_band, num_band)):
            bi = self._band_indices[i]
            if f1[bi] > cutoff and f2[j] > cutoff and f3[k] > cutoff:
                fc3_elem = self._sum_in_atoms((bi, j, k), (e1, e2, e3))
                fff = np.sqrt(f1[bi] * f2[j] * f3[k])
                self._fc3_normal[i, j, k] = fc3_elem / fff

    def _sum_in_atoms(self, band_indices, eigvecs):
        num_atom = self._primitive.get_number_of_atoms()
        (e1, e2, e3) = eigvecs
        (b1, b2, b3) = band_indices

        sum_fc3 = 0j
        for (i, j, k) in list(np.ndindex((num_atom,) * 3)):
            sum_fc3_cart = 0
            for (l, m, n) in list(np.ndindex((3, 3, 3))):
                sum_fc3_cart += (e1[i * 3 + l, b1] *
                                 e2[j * 3 + m, b2] *
                                 e3[k * 3 + n, b3] *
                                 self._fc3_reciprocal[i, j, k, l, m, n])
            mass_sqrt = np.sqrt(np.prod(self._masses[[i, j, k]]))
            sum_fc3 += sum_fc3_cart / mass_sqrt

        return sum_fc3
